! WHIZARD 2.0.0 Mon Aug 24 2009
! 
! (C) 1999-2009 by 
!     Wolfgang Kilian <kilian@hep.physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@physik.uni-freiburg.de>
!     with contributions by Sebastian Schmidt
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module evaluators

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils
  use lorentz
  use models
  use flavors
  use colors
  use helicities
  use quantum_numbers
  use state_matrices
  use interactions

  implicit none
  private

  public :: evaluator_t
  public :: evaluator_write
  public :: assignment(=)
  public :: evaluator_init_product
  public :: evaluator_init_square_with_color_factors
  public :: evaluator_init_squared_flows
  public :: evaluator_get_int_ptr
  public :: evaluator_is_empty
  public :: evaluator_get_n_tot
  public :: evaluator_get_n_in
  public :: evaluator_get_n_vir
  public :: evaluator_get_n_out
  public :: evaluator_sum
  public :: evaluator_get_mask
  public :: interaction_set_source_link
  public :: evaluator_set_source_link
  public :: evaluator_receive_momenta
  public :: evaluator_reassign_links
  public :: evaluator_get_unstable_particle
  public :: evaluator_final
  public :: evaluator_evaluate
  public :: evaluator_test

  integer, parameter :: &
       EVAL_UNDEFINED = 0, &
       EVAL_PRODUCT = 1, &
       EVAL_SQUARED_FLOWS = 2, &
       EVAL_SQUARE_WITH_COLOR_FACTORS = 3

  type :: pairing_array_t
     integer, dimension(:), allocatable :: i1, i2
     complex(default), dimension(:), allocatable :: factor
  end type pairing_array_t
     
  type :: evaluator_t
     private
     integer :: type = EVAL_UNDEFINED
     type(interaction_t), pointer :: int_in1 => null ()
     type(interaction_t), pointer :: int_in2 => null ()
     type(interaction_t) :: int
     type(pairing_array_t), dimension(:), allocatable :: pairing_array
  end type evaluator_t


  interface assignment(=)
     module procedure evaluator_assign
  end interface

  interface evaluator_init_product
     module procedure evaluator_init_product_ii
     module procedure evaluator_init_product_ie
     module procedure evaluator_init_product_ei
     module procedure evaluator_init_product_ee
  end interface

  interface interaction_set_source_link
     module procedure interaction_set_source_link_eval
  end interface
  interface evaluator_set_source_link
     module procedure evaluator_set_source_link_int
     module procedure evaluator_set_source_link_eval
  end interface
  interface evaluator_reassign_links
     module procedure evaluator_reassign_links_eval
     module procedure evaluator_reassign_links_int
  end interface


contains

  elemental subroutine pairing_array_init (pa, n, has_i2, has_factor)
    type(pairing_array_t), intent(out) :: pa
    integer, intent(in) :: n
    logical, intent(in) :: has_i2, has_factor
    allocate (pa%i1 (n))
    if (has_i2)  allocate (pa%i2 (n))
    if (has_factor)  allocate (pa%factor (n))
  end subroutine pairing_array_init

  subroutine evaluator_write (eval, unit, verbose, show_momentum_sum, show_mass)
    type(evaluator_t), intent(in) :: eval
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: verbose, show_momentum_sum, show_mass
    logical :: conjugate
    integer :: u, i, j
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(A)")  "Evaluator:"
    call interaction_write &
         (eval%int, unit, verbose, show_momentum_sum, show_mass)
    write (u, "(1x,A)")  "Matrix-element multiplication"
    write (u, "(2x,A)", advance="no")  "Input interaction 1:"
    if (associated (eval%int_in1)) then
       write (u, "(1x,I0)")  interaction_get_tag (eval%int_in1)
    else
       write (u, *)  " [undefined]"
    end if
    write (u, "(2x,A)", advance="no")  "Input interaction 2:"
    if (associated (eval%int_in2)) then
       write (u, *)  interaction_get_tag (eval%int_in2)
    else
       write (u, *)  " [undefined]"
    end if
    select case (eval%type)
    case (EVAL_SQUARED_FLOWS, EVAL_SQUARE_WITH_COLOR_FACTORS)
       conjugate = .true.
    case default
       conjugate = .false.
    end select
    if (allocated (eval%pairing_array)) then
       do i = 1, size (eval%pairing_array)
          write (u, "(2x,A,I0,A)")  "ME(", i, ") = "
          do j = 1, size (eval%pairing_array(i)%i1)
             write (u, "(4x,A)", advance="no")  "+"
             if (allocated (eval%pairing_array(i)%i2)) then
                write (u, "(1x,A,I0,A)", advance="no")  &
                  "ME1(", eval%pairing_array(i)%i1(j), ")"
                if (conjugate) then
                   write (u, "(A)", advance="no")  "* x"
                else
                   write (u, "(A)", advance="no")  " x"
                end if
                write (u, "(1x,A,I0,A)", advance="no")  &
                  "ME2(", eval%pairing_array(i)%i2(j), ")"
             else
                write (u, "(1x,A)", advance="no")  "|"
                write (u, "(A,I0,A)", advance="no")  &
                  "ME1(", eval%pairing_array(i)%i1(j), ")"
                write (u, "(A)", advance="no")  "|^2"
             end if
             if (allocated (eval%pairing_array(i)%factor)) then
                write (u, "(1x,A)", advance="no")  "x"
                write (u, *)  eval%pairing_array(i)%factor(j)
             else
                write (u, *)
             end if
          end do
       end do
    end if
  end subroutine evaluator_write

  subroutine evaluator_assign (eval_out, eval_in)
    type(evaluator_t), intent(out) :: eval_out
    type(evaluator_t), intent(in) :: eval_in
    eval_out%type = eval_in%type
    eval_out%int_in1 => eval_in%int_in1
    eval_out%int_in2 => eval_in%int_in2
    eval_out%int = eval_in%int
    if (allocated (eval_in%pairing_array)) then
       allocate (eval_out%pairing_array (size (eval_in%pairing_array)))
       eval_out%pairing_array = eval_in%pairing_array
    end if
  end subroutine evaluator_assign

  subroutine evaluator_init_product_ii &
       (eval, int_in1, int_in2, qn_mask_conn, qn_mask_rest, &
        connections_are_resonant)
    type(evaluator_t), intent(out), target :: eval
    type(interaction_t), intent(in), target :: int_in1, int_in2
    type(quantum_numbers_mask_t), intent(in) :: qn_mask_conn
    type(quantum_numbers_mask_t), intent(in), optional :: qn_mask_rest
    logical, intent(in), optional :: connections_are_resonant
    integer, dimension(:,:), allocatable :: connection_index
    integer :: n_in, n_vir, n_out, n_tot
    integer :: n1, n2, n_connections, n_me_connections
    integer :: color_max1, color_max_conn
    integer, dimension(:), allocatable :: map1, map2, map_connections
    integer, dimension(:), allocatable :: me_index1, me_index2
    logical, dimension(:), allocatable :: connection_mask1, connection_mask2
    type(quantum_numbers_mask_t), dimension(:), allocatable :: &
         qn_mask_conn_initial
    type(state_matrix_t) :: state_connections
    type :: connection_table_t
       type(quantum_numbers_t), dimension(:), allocatable :: qn_conn
       integer, dimension(2) :: n_index = 0
       integer, dimension(:), allocatable :: index1, index2
       type(quantum_numbers_t), dimension(:,:), allocatable :: qn1, qn2
    end type connection_table_t
    type(connection_table_t), dimension(:), allocatable :: connection_table
    integer, dimension(:), allocatable :: result_index
    eval%type = EVAL_PRODUCT
    eval%int_in1 => int_in1
    eval%int_in2 => int_in2
!     print *, "Evaluator product"
!     print *, "First interaction"
!     call interaction_write (int_in1)
!     print *
!     print *, "Second interaction"
!     call interaction_write (int_in2)
!     print *
    color_max1 = interaction_get_max_color_value (int_in1)
    ! color_max2 = interaction_get_max_color_value (int_in2)
    call find_connections (int_in1, int_in2, n_connections, connection_index)
    call compute_index_bounds_and_mappings &
         (int_in1, int_in2, n_connections, &
          n_in, n_vir, n_out, n_tot, &
          n1, n2, map1, map2, map_connections)
    allocate (me_index1 (interaction_get_n_matrix_elements (int_in1)))
    allocate (me_index2 (interaction_get_n_matrix_elements (int_in2)))
    allocate (connection_mask1 (interaction_get_n_tot (int_in1)))
    allocate (connection_mask2 (interaction_get_n_tot (int_in2)))
    connection_mask1 = .true.
    connection_mask2 = .true.
    connection_mask1(connection_index(:,1)) = .false.
    connection_mask2(connection_index(:,2)) = .false.
    allocate (qn_mask_conn_initial (n_connections))
    qn_mask_conn_initial = &
         interaction_get_mask (int_in1, connection_index(:,1)) .or. &
         interaction_get_mask (int_in2, connection_index(:,2))
    call state_matrix_init (state_connections, n_counters=2)
    call accumulate_connected_states (state_connections, me_index1, &
         interaction_get_state_matrix (int_in1), &
         connection_index(:,1), qn_mask_conn_initial, n_connections, 1)
    call accumulate_connected_states (state_connections, me_index2, &
         interaction_get_state_matrix (int_in2), &
         connection_index(:,2), qn_mask_conn_initial, n_connections, 2)
    color_max_conn = state_matrix_get_max_color_value (state_connections)
    n_me_connections = state_matrix_get_n_matrix_elements (state_connections)
    allocate (connection_table (n_me_connections))
    call allocate_connection_entries &
         (connection_table, state_connections, n_connections, n1, n2)
    call fill_connection_table (connection_table, &
         interaction_get_state_matrix (int_in1), &
         me_index1, connection_mask1, connection_index(:,1), 1, &
         color_max_conn)
    call fill_connection_table (connection_table, &
         interaction_get_state_matrix (int_in2), &
         me_index2, connection_mask2, connection_index(:,2), 2, &
         color_max_conn + color_max1)
    call make_product_interaction (eval%int, &
         result_index, &
         int_in1, int_in2, &
         n_in, n_vir, n_out, &
         connection_table, n1, n2, n_connections, &
         connection_mask1, connection_mask2, &
         qn_mask_conn_initial, qn_mask_conn, qn_mask_rest)
    call make_pairing_array (eval%pairing_array, &
         interaction_get_n_matrix_elements (eval%int), &
         result_index, connection_table)
    call record_links (eval%int, &
         int_in1, int_in2, connection_index, map1, map2, map_connections, &
         connection_mask1, connection_mask2, connections_are_resonant)
    call state_matrix_final (state_connections)
!     print *, "Result evaluator"
!     call evaluator_write (eval)

  contains

    subroutine compute_index_bounds_and_mappings &
         (int1, int2, n_connections, &
          n_in, n_vir, n_out, n_tot, &
          n_tot1, n_tot2, map1, map2, map_connections)
      type(interaction_t), intent(in) :: int1, int2
      integer, intent(in) :: n_connections
      integer, intent(out) :: n_in, n_vir, n_out, n_tot
      integer, intent(out) :: n_tot1, n_tot2
      integer, dimension(:), allocatable, intent(out) :: map1, map2
      integer, dimension(:), allocatable, intent(out) :: map_connections
      integer, dimension(:), allocatable :: index
      integer :: n_in1, n_vir1, n_out1
      integer :: n_in2, n_vir2, n_out2
      integer :: k, i
      n_in1  = interaction_get_n_in  (int1)
      n_vir1 = interaction_get_n_vir (int1)
      n_out1 = interaction_get_n_out (int1) - n_connections
      n_tot1 = n_in1 + n_vir1 + n_out1
      n_in2  = interaction_get_n_in  (int2) - n_connections
      n_vir2 = interaction_get_n_vir (int2)
      n_out2 = interaction_get_n_out (int2)
      n_tot2 = n_in2 + n_vir2 + n_out2
      n_in  = n_in1  + n_in2
      n_vir = n_vir1 + n_vir2 + n_connections
      n_out = n_out1 + n_out2
      n_tot = n_in + n_vir + n_out
      allocate (map1 (n_tot1))
      allocate (map2 (n_tot2))
      allocate (map_connections (n_connections))
      allocate (index (n_tot))
      index = (/ (i, i = 1, n_tot) /)
      map1(1 : n_in1) = index(  1 :   n_in1);  k =     n_in1
      map2(1 : n_in2) = index(k+1 : k+n_in2);  k = k + n_in2
      map1(n_in1+1 : n_in1+n_vir1) = index(k+1 : k+n_vir1);  k = k + n_vir1
      map2(n_in2+1 : n_in2+n_vir2) = index(k+1 : k+n_vir2);  k = k + n_vir2
      map_connections = index(k+1 : k+n_connections);  k = k + n_connections
      map1(n_in1+n_vir1+1 : n_tot1) = index(k+1 : k+n_out1);  k = k + n_out1
      map2(n_in2+n_vir2+1 : n_tot2) = index(k+1 : k+n_out2)
    end subroutine compute_index_bounds_and_mappings

    subroutine accumulate_connected_states (state_connections, me_index, &
         state, connection_index, qn_mask_conn, n_connections, i)
      type(state_matrix_t), intent(inout) :: state_connections
      integer, dimension(:), intent(out) :: me_index
      type(state_matrix_t), intent(in) :: state
      integer, dimension(:), intent(in) :: connection_index
      type(quantum_numbers_mask_t), dimension(:), intent(in) :: qn_mask_conn
      integer, intent(in) :: n_connections
      integer, intent(in) :: i
      type(state_iterator_t) :: it
      type(quantum_numbers_t), dimension(n_connections) :: qn
      integer :: me_index_state
      call state_iterator_init (it, state)
      do while (state_iterator_is_valid (it))
         qn = state_iterator_get_quantum_numbers (it, connection_index)
         call quantum_numbers_undefine (qn, qn_mask_conn)
         call quantum_numbers_canonicalize_color (qn)
         me_index_state = state_iterator_get_me_index (it)
         call state_matrix_add_state (state_connections, qn, &
              counter_index = i, &
              me_index = me_index(me_index_state))
         call state_iterator_advance (it)
      end do
    end subroutine accumulate_connected_states

    subroutine allocate_connection_entries &
         (connection_table, state_connections, n_connections, len1, len2)
      type(connection_table_t), dimension(:), intent(inout) :: &
           connection_table
      type(state_matrix_t), intent(in) :: state_connections
      integer, intent(in) :: n_connections, len1, len2
      type(state_iterator_t) :: it
      integer :: i
      integer, dimension(2) :: n
      call state_iterator_init (it, state_connections)
      do while (state_iterator_is_valid (it))
         i = state_iterator_get_me_index (it)
         n = state_iterator_get_me_count (it)
         allocate (connection_table(i)%qn_conn (n_connections))
         connection_table(i)%qn_conn = state_iterator_get_quantum_numbers (it)
         connection_table(i)%n_index = n
         allocate (connection_table(i)%index1 (n(1)))
         allocate (connection_table(i)%index2 (n(2)))
         allocate (connection_table(i)%qn1 (len1, n(1)))
         allocate (connection_table(i)%qn2 (len2, n(2)))
         call state_iterator_advance (it)
      end do
    end subroutine allocate_connection_entries

    subroutine fill_connection_table (connection_table, &
         state, me_index, connection_mask, connection_index, i, offset)
      type(connection_table_t), dimension(:), intent(inout) :: &
           connection_table
      type(state_matrix_t), intent(in) :: state
      integer, dimension(:), intent(in) :: me_index
      logical, dimension(:), intent(in) :: connection_mask
      integer, dimension(:), intent(in) :: connection_index
      integer, dimension(:,:), allocatable :: color_map
      integer, intent(in) :: i, offset
      integer :: k, index_state, index_conn
      integer, dimension(size(connection_table)) :: count_conn
      type(quantum_numbers_t), dimension(:), allocatable :: qn
      type(state_iterator_t) :: it
      allocate (qn (state_matrix_get_depth (state)))
      count_conn = 0
      call state_iterator_init (it, state)
      do while (state_iterator_is_valid (it))
         index_state = state_iterator_get_me_index (it)
         index_conn = me_index(index_state)
         count_conn(index_conn) = count_conn(index_conn) + 1
         k = count_conn(index_conn)
         qn = state_iterator_get_quantum_numbers (it)
         call quantum_numbers_set_color_map (color_map, &
              qn(connection_index), connection_table(index_conn)%qn_conn)
         select case (i)
         case (1)
            connection_table(index_conn)%index1(k) = index_state
            connection_table(index_conn)%qn1(:,k) = pack (qn, connection_mask)
            call quantum_numbers_translate_color &
                 (connection_table(index_conn)%qn1(:,k), color_map, offset)
         case (2)
            connection_table(index_conn)%index2(k) = index_state
            connection_table(index_conn)%qn2(:,k) = pack (qn, connection_mask)
            call quantum_numbers_translate_color &
                 (connection_table(index_conn)%qn2(:,k), color_map, offset)
         end select
         call state_iterator_advance (it)
      end do
    end subroutine fill_connection_table

    subroutine make_product_interaction (int, &
         result_index, &
         int1, int2, &
         n_in, n_vir, n_out, &
         connection_table, n1, n2, n_connections, &
         connection_mask1, connection_mask2, &
         qn_mask_conn_initial, qn_mask_conn, qn_mask_rest)
      type(interaction_t), intent(out), target :: int
      integer, dimension(:), intent(out), allocatable :: result_index
      type(interaction_t), intent(in) :: int1, int2
      integer, intent(in) :: n_in, n_vir, n_out
      type(connection_table_t), dimension(:), intent(in) :: connection_table
      integer, intent(in) :: n1, n2, n_connections
      type(quantum_numbers_mask_t), dimension(:), intent(in) :: &
           qn_mask_conn_initial
      type(quantum_numbers_mask_t), intent(in) :: qn_mask_conn
      type(quantum_numbers_mask_t), intent(in), optional :: qn_mask_rest
      logical, dimension(:), intent(in) :: connection_mask1, connection_mask2
      integer, dimension(n1) :: index1
      integer, dimension(n2) :: index2
      integer, dimension(n_connections) :: index_conn
      integer :: i, j, k, m
      integer :: n_result_entries
      type(quantum_numbers_t), dimension(n1+n2+n_connections) :: qn
      type(quantum_numbers_mask_t), dimension(n1+n2+n_connections) :: qn_mask
      index1 = map1 ((/ (i, i = 1, n1) /))
      index2 = map2 ((/ (i, i = 1, n2) /))
      index_conn = map_connections ((/ (i, i = 1, n_connections) /))
      if (present (qn_mask_rest)) then
         qn_mask(index1) = &
              pack (interaction_get_mask (int1), connection_mask1) &
              .or. qn_mask_rest
         qn_mask(index2) = &
              pack (interaction_get_mask (int2), connection_mask2) &
              .or. qn_mask_rest
      else
         qn_mask(index1) = &
              pack (interaction_get_mask (int1), connection_mask1)
         qn_mask(index2) = &
              pack (interaction_get_mask (int2), connection_mask2)
      end if
      qn_mask(index_conn) = qn_mask_conn_initial .or. qn_mask_conn
      call interaction_init (eval%int, n_in, n_vir, n_out, mask=qn_mask)
      n_result_entries = &
           sum (connection_table%n_index(1) * connection_table%n_index(2))
      allocate (result_index (n_result_entries))
      m = 1
      do i = 1, size (connection_table)
         qn(index_conn) = quantum_numbers_undefined &
              (connection_table(i)%qn_conn, qn_mask_conn)
         do j = 1, connection_table(i)%n_index(1)
            qn(index1) = connection_table(i)%qn1(:,j)
            do k = 1, connection_table(i)%n_index(2)
               qn(index2) = connection_table(i)%qn2(:,k)
               call interaction_add_state (int, qn, &
                    me_index = result_index(m))
               m = m + 1
            end do
         end do
      end do
      call interaction_set_mask (int, qn_mask)
      call interaction_freeze (int)
    end subroutine make_product_interaction

    subroutine make_pairing_array (pa, &
         n_matrix_elements, result_index, connection_table)
      type(pairing_array_t), dimension(:), intent(out), allocatable :: pa
      integer, intent(in) :: n_matrix_elements
      integer, dimension(:), intent(in) :: result_index
      type(connection_table_t), dimension(:), intent(in) :: connection_table
      integer, dimension(:), allocatable :: n_entries
      integer :: i, j, k, m, r
      allocate (pa (n_matrix_elements))
      allocate (n_entries (n_matrix_elements))
      n_entries = 0
      do m = 1, size (result_index)
         r = result_index(m)
         n_entries(r) = n_entries(r) + 1
      end do
      call pairing_array_init &
           (pa, n_entries, has_i2=.true., has_factor=.false.)
      m = 1
      n_entries = 0
      do i = 1, size (connection_table)
         do j = 1, connection_table(i)%n_index(1)
            do k = 1, connection_table(i)%n_index(2)
               r = result_index(m)
               n_entries(r) = n_entries(r) + 1
               pa(r)%i1(n_entries(r)) = connection_table(i)%index1(j)
               pa(r)%i2(n_entries(r)) = connection_table(i)%index2(k)
               m = m + 1
            end do
         end do
      end do
    end subroutine make_pairing_array

    subroutine record_links (int, &
         int_in1, int_in2, connection_index, map1, map2, map_connections, &
         connection_mask1, connection_mask2, resonant)
      type(interaction_t), intent(inout) :: int
      type(interaction_t), intent(in), target :: int_in1, int_in2
      integer, dimension(:,:), intent(in) :: connection_index
      integer, dimension(:), intent(in) :: map1, map2, map_connections
      logical, dimension(:), intent(in) :: connection_mask1, connection_mask2
      logical, intent(in), optional :: resonant
      integer, dimension(:), allocatable :: map_all1, map_all2
      integer :: i, j, k
      allocate (map_all1 (size (connection_mask1)))
      k = 0
      j = 0
      do i = 1, size (connection_mask1)
         if (connection_mask1(i)) then
            j = j + 1
            map_all1(i) = map1(j)
         else
            k = k + 1
            map_all1(i) = map_connections(k)
         end if
         call interaction_set_source_link (int, map_all1(i), int_in1, i)
      end do
      call interaction_transfer_relations (int_in1, int, map_all1)
      allocate (map_all2 (size (connection_mask2)))
      j = 0
      do i = 1, size (connection_mask2)
         if (connection_mask2(i)) then
            j = j + 1
            map_all2(i) = map2(j)
            call interaction_set_source_link (int, map_all2(i), int_in2, i)
         else
            map_all2(i) = 0
         end if
      end do
      call interaction_transfer_relations (int_in2, int, map_all2)
      call interaction_relate_connections (int, &
           int_in2, connection_index(:,2), map_all2, map_connections, resonant)
    end subroutine record_links

  end subroutine evaluator_init_product_ii

  subroutine evaluator_init_product_ie &
       (eval, int_in1, eval_in2, qn_mask_conn, qn_mask_rest, &
        connections_are_resonant)
    type(evaluator_t), intent(out), target :: eval
    type(interaction_t), intent(in), target :: int_in1
    type(evaluator_t), intent(in), target :: eval_in2
    type(quantum_numbers_mask_t), intent(in) :: qn_mask_conn
    type(quantum_numbers_mask_t), intent(in), optional :: qn_mask_rest
    logical, intent(in), optional :: connections_are_resonant
    call evaluator_init_product_ii &
         (eval, int_in1, eval_in2%int, qn_mask_conn, qn_mask_rest, &
          connections_are_resonant)
  end subroutine evaluator_init_product_ie

  subroutine evaluator_init_product_ei &
       (eval, eval_in1, int_in2, qn_mask_conn, qn_mask_rest, &
        connections_are_resonant)
    type(evaluator_t), intent(out), target :: eval
    type(evaluator_t), intent(in), target :: eval_in1
    type(interaction_t), intent(in), target :: int_in2
    type(quantum_numbers_mask_t), intent(in) :: qn_mask_conn
    type(quantum_numbers_mask_t), intent(in), optional :: qn_mask_rest
    logical, intent(in), optional :: connections_are_resonant
    call evaluator_init_product_ii &
         (eval, eval_in1%int, int_in2, qn_mask_conn, qn_mask_rest, &
         connections_are_resonant)
  end subroutine evaluator_init_product_ei

  subroutine evaluator_init_product_ee &
       (eval, eval_in1, eval_in2, qn_mask_conn, qn_mask_rest, &
        connections_are_resonant)
    type(evaluator_t), intent(out), target :: eval
    type(evaluator_t), intent(in), target :: eval_in1, eval_in2
    type(quantum_numbers_mask_t), intent(in) :: qn_mask_conn
    type(quantum_numbers_mask_t), intent(in), optional :: qn_mask_rest
    logical, intent(in), optional :: connections_are_resonant
    call evaluator_init_product_ii &
         (eval, eval_in1%int, eval_in2%int, qn_mask_conn, qn_mask_rest, &
          connections_are_resonant)
  end subroutine evaluator_init_product_ee

  subroutine evaluator_init_square_with_color_factors &
       (eval, int_in, qn_mask, nc)
    type(evaluator_t), intent(out), target :: eval
    type(interaction_t), intent(in), target :: int_in
    type(quantum_numbers_mask_t), dimension(:), intent(in) :: qn_mask
    integer, intent(in), optional :: nc
    integer :: ncol
    integer :: n_in, n_vir, n_out, n_tot
    integer, dimension(:), allocatable :: me_index
    type(quantum_numbers_mask_t), dimension(:), allocatable :: qn_mask_initial
    type(state_matrix_t) :: state_connections
    integer :: n_me_connections
    type :: connection_table_t
       type(quantum_numbers_t), dimension(:), allocatable :: qn_conn
       integer :: n_index = 0
       integer, dimension(:), allocatable :: index
       type(quantum_numbers_t), dimension(:,:), allocatable :: qn
    end type connection_table_t
    type(connection_table_t), dimension(:), allocatable :: connection_table
    integer, dimension(:), allocatable :: result_index
    eval%type = EVAL_SQUARE_WITH_COLOR_FACTORS
    ncol = 3;  if (present (nc))  ncol = nc
    eval%int_in1 => int_in
!     print *, "Interaction square with color factors"
!     print *, "Input interaction"
!     call interaction_write (int_in)
    n_in  = interaction_get_n_in  (int_in)
    n_vir = interaction_get_n_vir (int_in)
    n_out = interaction_get_n_out (int_in)
    n_tot = interaction_get_n_tot (int_in)
    allocate (me_index (interaction_get_n_matrix_elements (int_in)))
    allocate (qn_mask_initial (n_tot))
    qn_mask_initial = interaction_get_mask (int_in)
    call quantum_numbers_mask_set_color (qn_mask_initial, &
         .true., mask_cg=.false.)
    call state_matrix_init (state_connections, n_counters=1)
    call accumulate_connected_states (state_connections, me_index, &
         interaction_get_state_matrix (int_in), qn_mask_initial, n_tot)
    n_me_connections = state_matrix_get_n_matrix_elements (state_connections)
    allocate (connection_table (n_me_connections))
    call allocate_connection_entries (connection_table, &
         state_connections, n_tot)
    call fill_connection_table (connection_table, &
         interaction_get_state_matrix (int_in), me_index)
    call make_squared_interaction (eval%int, &
         result_index, &
         n_in, n_vir, n_out, n_tot, &
         connection_table, qn_mask_initial .or. qn_mask)
    call make_pairing_array (eval%pairing_array, &
         interaction_get_n_matrix_elements (eval%int), &
         result_index, connection_table, n_in, n_tot, ncol)
    call record_links (eval%int, int_in, n_tot)
    call state_matrix_final (state_connections)
!     print *, "Result evaluator:"
!     call evaluator_write (eval)

  contains
    
    subroutine accumulate_connected_states &
         (state_connections, me_index, state, qn_mask, n_tot)
      type(state_matrix_t), intent(inout) :: state_connections
      integer, dimension(:), intent(out) :: me_index
      type(state_matrix_t), intent(in) :: state
      type(quantum_numbers_mask_t), dimension(:), intent(in) :: qn_mask
      integer, intent(in) :: n_tot
      type(quantum_numbers_t), dimension(n_tot) :: qn
      type(state_iterator_t) :: it
      integer :: me_index_state
      call state_iterator_init (it, state)
      do while (state_iterator_is_valid (it))
         qn = state_iterator_get_quantum_numbers (it)
         call quantum_numbers_undefine (qn, qn_mask)
         me_index_state = state_iterator_get_me_index (it)
         call state_matrix_add_state (state_connections, qn, &
              counter_index = 1, &
              me_index = me_index(me_index_state))
         call state_iterator_advance (it)
      end do
    end subroutine accumulate_connected_states
         
    subroutine allocate_connection_entries &
         (connection_table, state_connections, n_tot)
      type(connection_table_t), dimension(:), intent(inout) :: &
           connection_table
      type(state_matrix_t), intent(in) :: state_connections
      integer, intent(in) :: n_tot
      type(state_iterator_t) :: it
      integer :: i
      integer, dimension(1) :: n
      call state_iterator_init (it, state_connections)
      do while (state_iterator_is_valid (it))
         i = state_iterator_get_me_index (it)
         n = state_iterator_get_me_count (it)
         allocate (connection_table(i)%qn_conn (n_tot))
         connection_table(i)%qn_conn = state_iterator_get_quantum_numbers (it)
         connection_table(i)%n_index = n(1)
         allocate (connection_table(i)%index (n(1)))
         allocate (connection_table(i)%qn (n_tot, n(1)))
         call state_iterator_advance (it)
      end do
    end subroutine allocate_connection_entries

    subroutine fill_connection_table (connection_table, state, me_index)
      type(connection_table_t), dimension(:), intent(inout) :: &
           connection_table
      type(state_matrix_t), intent(in) :: state
      integer, dimension(:), intent(in) :: me_index
      integer :: k, index_state, index_conn
      integer, dimension(size(connection_table)) :: count_conn
      type(state_iterator_t) :: it
      count_conn = 0
      call state_iterator_init (it, state)
      do while (state_iterator_is_valid (it))
         index_state = state_iterator_get_me_index (it)
         index_conn = me_index(index_state)
         count_conn(index_conn) = count_conn(index_conn) + 1
         k = count_conn(index_conn)
         connection_table(index_conn)%index(k) = index_state
         connection_table(index_conn)%qn(:,k) = &
              state_iterator_get_quantum_numbers (it)
         call state_iterator_advance (it)
      end do
    end subroutine fill_connection_table

    subroutine make_squared_interaction (int, &
         result_index, &
         n_in, n_vir, n_out, n_tot, &
         connection_table, qn_mask)
      type(interaction_t), intent(out), target :: int
      integer, dimension(:), intent(out), allocatable :: result_index
      integer, intent(in) :: n_in, n_vir, n_out, n_tot
      type(connection_table_t), dimension(:), intent(in) :: connection_table
      type(quantum_numbers_mask_t), dimension(:), intent(in) :: qn_mask
      integer :: i, k, m
      integer :: n_result_entries
      type(quantum_numbers_t), dimension(n_tot) :: qn
      call interaction_init (eval%int, n_in, n_vir, n_out, mask=qn_mask)
      n_result_entries = sum (connection_table%n_index ** 2)
      allocate (result_index (n_result_entries))
      m = 1
      do i = 1, size (connection_table)
         qn = quantum_numbers_undefined (connection_table(i)%qn_conn, qn_mask)
         call interaction_add_state (int, qn, me_index = result_index(m))
         k = connection_table(i)%n_index ** 2
         result_index(m+1:m+k-1) = result_index(m)
         m = m + k
      end do
      call interaction_set_mask (int, qn_mask)
      call interaction_freeze (int)
    end subroutine make_squared_interaction

    subroutine make_pairing_array (pa, &
         n_matrix_elements, result_index, connection_table, n_in, n_tot, ncol)
      type(pairing_array_t), dimension(:), intent(out), allocatable :: pa
      integer, intent(in) :: n_matrix_elements
      integer, dimension(:), intent(in) :: result_index
      type(connection_table_t), dimension(:), intent(in) :: connection_table
      integer, intent(in) :: n_in, n_tot, ncol
      integer, dimension(:), allocatable :: n_entries
      integer :: i, j, k, m, r
      integer :: nloops, nghost
      real(default) :: factor
      type(quantum_numbers_t), dimension(n_tot) :: qn
      allocate (pa (n_matrix_elements))
      allocate (n_entries (n_matrix_elements))
      n_entries = 0
      do m = 1, size (result_index)
         r = result_index(m)
         n_entries(r) = n_entries(r) + 1
      end do
      call pairing_array_init &
           (pa, n_entries, has_i2=.true., has_factor=.true.)
      m = 1
      n_entries = 0
      do i = 1, size (connection_table)
         do j = 1, connection_table(i)%n_index
            do k = 1, connection_table(i)%n_index
               r = result_index(m)
               n_entries(r) = n_entries(r) + 1
               pa(r)%i1(n_entries(r)) = connection_table(i)%index(j)
               pa(r)%i2(n_entries(r)) = connection_table(i)%index(k)
               qn = connection_table(i)%qn(:,j) .merge. &
                    connection_table(i)%qn(:,k)
               nloops = count_color_loops (quantum_numbers_get_color (qn))
               nghost = count (quantum_numbers_is_color_ghost (qn))
               factor = real (ncol, default) ** (nloops - nghost) &
                    / product (abs (quantum_numbers_get_color_type (qn(:n_in))))
               if (quantum_numbers_ghost_parity (qn))  factor = - factor
               pa(r)%factor(n_entries(r)) = factor
               m = m + 1
            end do
         end do
      end do
    end subroutine make_pairing_array

    subroutine record_links (int, int_in, n_tot)
      type(interaction_t), intent(inout) :: int
      type(interaction_t), intent(in), target :: int_in
      integer, intent(in) :: n_tot
      integer, dimension(n_tot) :: map
      integer :: i
      do i = 1, n_tot
         call interaction_set_source_link (int, i, int_in, i)
      end do
      map = (/ (i, i = 1, n_tot) /)
      call interaction_transfer_relations (int_in, int, map)
    end subroutine record_links

  end subroutine evaluator_init_square_with_color_factors

  subroutine evaluator_init_squared_flows (eval, int_in, qn_mask)
    type(evaluator_t), intent(out), target :: eval
    type(interaction_t), intent(in), target :: int_in
    type(quantum_numbers_mask_t), dimension(:), intent(in) :: qn_mask
    integer :: n_in, n_vir, n_out, n_tot
    integer, dimension(:), allocatable :: me_index
    type(quantum_numbers_mask_t), dimension(:), allocatable :: qn_mask_initial
    type(state_matrix_t) :: state_connections
    integer :: n_me_connections
    type :: connection_table_t
       type(quantum_numbers_t), dimension(:), allocatable :: qn_conn
       integer :: index
    end type connection_table_t
    type(connection_table_t), dimension(:), allocatable :: connection_table
    integer, dimension(:), allocatable :: result_index
    eval%type = EVAL_SQUARED_FLOWS
    eval%int_in1 => int_in
!     print *, "Interaction squared as flow coefficients"
!     print *, "Input interaction"
!     call interaction_write (int_in)
    n_in  = interaction_get_n_in  (int_in)
    n_vir = interaction_get_n_vir (int_in)
    n_out = interaction_get_n_out (int_in)
    n_tot = interaction_get_n_tot (int_in)
    allocate (me_index (interaction_get_n_matrix_elements (int_in)))
    allocate (qn_mask_initial (n_tot))
    qn_mask_initial = interaction_get_mask (int_in)
    call quantum_numbers_mask_set_color (qn_mask_initial, .false.)
    call state_matrix_init (state_connections, n_counters=1)
    call accumulate_connected_states (state_connections, me_index, &
         interaction_get_state_matrix (int_in), qn_mask_initial, n_tot)
    n_me_connections = state_matrix_get_n_matrix_elements (state_connections)
    allocate (connection_table (n_me_connections))
    call allocate_connection_entries (connection_table, &
         state_connections, n_tot)
    call fill_connection_table (connection_table, &
         interaction_get_state_matrix (int_in), me_index)
    call make_squared_interaction (eval%int, &
         result_index, &
         n_in, n_vir, n_out, n_tot, &
         connection_table, qn_mask_initial .or. qn_mask)
    call make_pairing_array (eval%pairing_array, &
         interaction_get_n_matrix_elements (eval%int), &
         result_index, connection_table, n_in, n_tot)
    call record_links (eval%int, int_in, n_tot)
    call state_matrix_final (state_connections)
!     print *, "Result evaluator:"
!     call evaluator_write (eval)

  contains
    
    subroutine accumulate_connected_states &
         (state_connections, me_index, state, qn_mask, n_tot)
      type(state_matrix_t), intent(inout) :: state_connections
      integer, dimension(:), intent(out) :: me_index
      type(state_matrix_t), intent(in) :: state
      type(quantum_numbers_mask_t), dimension(:), intent(in) :: qn_mask
      integer, intent(in) :: n_tot
      type(quantum_numbers_t), dimension(n_tot) :: qn
      type(state_iterator_t) :: it
      integer :: me_index_state
      call state_iterator_init (it, state)
      do while (state_iterator_is_valid (it))
         qn = state_iterator_get_quantum_numbers (it)
         if (all (quantum_numbers_are_physical (qn))) then
            call quantum_numbers_undefine (qn, qn_mask)
            me_index_state = state_iterator_get_me_index (it)
            call state_matrix_add_state (state_connections, qn, &
                 me_index = me_index(me_index_state))
         end if
         call state_iterator_advance (it)
      end do
    end subroutine accumulate_connected_states
         
    subroutine allocate_connection_entries &
         (connection_table, state_connections, n_tot)
      type(connection_table_t), dimension(:), intent(inout) :: &
           connection_table
      type(state_matrix_t), intent(in) :: state_connections
      integer, intent(in) :: n_tot
      type(state_iterator_t) :: it
      integer :: i
      call state_iterator_init (it, state_connections)
      do while (state_iterator_is_valid (it))
         i = state_iterator_get_me_index (it)
         allocate (connection_table(i)%qn_conn (n_tot))
         connection_table(i)%qn_conn = state_iterator_get_quantum_numbers (it)
         call state_iterator_advance (it)
      end do
    end subroutine allocate_connection_entries

    subroutine fill_connection_table (connection_table, state, me_index)
      type(connection_table_t), dimension(:), intent(inout) :: &
           connection_table
      type(state_matrix_t), intent(in) :: state
      integer, dimension(:), intent(in) :: me_index
      integer :: index_state, index_conn
      type(state_iterator_t) :: it
      call state_iterator_init (it, state)
      do while (state_iterator_is_valid (it))
         index_state = state_iterator_get_me_index (it)
         index_conn = me_index(index_state)
         connection_table(index_conn)%index = index_state
         call state_iterator_advance (it)
      end do
    end subroutine fill_connection_table

    subroutine make_squared_interaction (int, &
         result_index, &
         n_in, n_vir, n_out, n_tot, &
         connection_table, qn_mask)
      type(interaction_t), intent(out), target :: int
      integer, dimension(:), intent(out), allocatable :: result_index
      integer, intent(in) :: n_in, n_vir, n_out, n_tot
      type(connection_table_t), dimension(:), intent(in) :: connection_table
      type(quantum_numbers_mask_t), dimension(:), intent(in) :: qn_mask
      integer :: i
      integer :: n_result_entries
      type(quantum_numbers_t), dimension(n_tot) :: qn
      call interaction_init (eval%int, n_in, n_vir, n_out, mask=qn_mask)
      n_result_entries = size (connection_table)
      allocate (result_index (n_result_entries))
      do i = 1, size (connection_table)
         qn = quantum_numbers_undefined (connection_table(i)%qn_conn, qn_mask)
         call interaction_add_state (int, qn, me_index = result_index(i))
      end do
      call interaction_set_mask (int, qn_mask)
      call interaction_freeze (int)
    end subroutine make_squared_interaction

    subroutine make_pairing_array (pa, &
         n_matrix_elements, result_index, connection_table, n_in, n_tot)
      type(pairing_array_t), dimension(:), intent(out), allocatable :: pa
      integer, intent(in) :: n_matrix_elements
      integer, dimension(:), intent(in) :: result_index
      type(connection_table_t), dimension(:), intent(in) :: connection_table
      integer, intent(in) :: n_in, n_tot
      integer, dimension(:), allocatable :: n_entries
      integer :: i, r
      allocate (pa (n_matrix_elements))
      allocate (n_entries (n_matrix_elements))
      n_entries = 0
      do i = 1, size (result_index)
         r = result_index(i)
         n_entries(r) = n_entries(r) + 1
      end do
      call pairing_array_init &
           (pa, n_entries, has_i2=.false., has_factor=.false.)
      n_entries = 0
      do i = 1, size (connection_table)
         r = result_index(i)
         n_entries(r) = n_entries(r) + 1
         pa(r)%i1(n_entries(r)) = connection_table(i)%index
      end do
    end subroutine make_pairing_array

    subroutine record_links (int, int_in, n_tot)
      type(interaction_t), intent(inout) :: int
      type(interaction_t), intent(in), target :: int_in
      integer, intent(in) :: n_tot
      integer, dimension(n_tot) :: map
      integer :: i
      do i = 1, n_tot
         call interaction_set_source_link (int, i, int_in, i)
      end do
      map = (/ (i, i = 1, n_tot) /)
      call interaction_transfer_relations (int_in, int, map)
    end subroutine record_links

  end subroutine evaluator_init_squared_flows

  function parity (mask)
    logical :: parity
    logical, dimension(:) :: mask
    integer :: i
    parity = .false.
    do i = 1, size (mask)
       if (mask(i))  parity = .not. parity
    end do
  end function parity

  function evaluator_get_int_ptr (eval) result (int)
    type(interaction_t), pointer :: int
    type(evaluator_t), intent(in), target :: eval
    int => eval%int
  end function evaluator_get_int_ptr

  function evaluator_is_empty (eval) result (flag)
    logical :: flag
    type(evaluator_t), intent(in) :: eval
    flag = interaction_is_empty (eval%int)
  end function evaluator_is_empty

  function evaluator_get_n_tot (eval) result (n_tot)
    integer :: n_tot
    type(evaluator_t), intent(in) :: eval
    n_tot = interaction_get_n_tot (eval%int)
  end function evaluator_get_n_tot

  function evaluator_get_n_in (eval) result (n_in)
    integer :: n_in
    type(evaluator_t), intent(in) :: eval
    n_in = interaction_get_n_in (eval%int)
  end function evaluator_get_n_in

  function evaluator_get_n_vir (eval) result (n_vir)
    integer :: n_vir
    type(evaluator_t), intent(in) :: eval
    n_vir = interaction_get_n_vir (eval%int)
  end function evaluator_get_n_vir

  function evaluator_get_n_out (eval) result (n_out)
    integer :: n_out
    type(evaluator_t), intent(in) :: eval
    n_out = interaction_get_n_out (eval%int)
  end function evaluator_get_n_out

  function evaluator_sum (eval) result (value)
    complex(default) :: value
    type(evaluator_t), intent(in) :: eval
    value = interaction_sum (eval%int)
  end function evaluator_sum

  function evaluator_get_mask (eval) result (mask)
    type(quantum_numbers_mask_t), dimension(:), allocatable :: mask
    type(evaluator_t), intent(in), target :: eval
    allocate (mask (interaction_get_n_tot (eval%int)))
    mask = interaction_get_mask (eval%int)
  end function evaluator_get_mask

  subroutine interaction_set_source_link_eval (int, i, eval1, i1)
    type(interaction_t), intent(inout) :: int
    type(evaluator_t), intent(in), target :: eval1
    integer, intent(in) :: i, i1
    call interaction_set_source_link (int, i, eval1%int, i1)
  end subroutine interaction_set_source_link_eval

  subroutine evaluator_set_source_link_int (eval, i, int1, i1)
    type(evaluator_t), intent(inout) :: eval
    type(interaction_t), intent(in), target :: int1
    integer, intent(in) :: i, i1
    call interaction_set_source_link (eval%int, i, int1, i1)
  end subroutine evaluator_set_source_link_int

  subroutine evaluator_set_source_link_eval (eval, i, eval1, i1)
    type(evaluator_t), intent(inout) :: eval
    type(evaluator_t), intent(in), target :: eval1
    integer, intent(in) :: i, i1
    call interaction_set_source_link (eval%int, i, eval1%int, i1)
  end subroutine evaluator_set_source_link_eval

  subroutine evaluator_receive_momenta (eval)
    type(evaluator_t), intent(inout) :: eval
    call interaction_receive_momenta (eval%int)
  end subroutine evaluator_receive_momenta

  subroutine evaluator_reassign_links_eval (eval, eval_src, eval_target)
    type(evaluator_t), intent(inout) :: eval
    type(evaluator_t), intent(in) :: eval_src
    type(evaluator_t), intent(in), target :: eval_target
    if (associated (eval%int_in1)) then
       if (interaction_get_tag (eval%int_in1) &
            == interaction_get_tag (eval_src%int)) then
          eval%int_in1 => eval_target%int
       end if
    end if
    if (associated (eval%int_in2)) then
       if (interaction_get_tag (eval%int_in2) &
            == interaction_get_tag (eval_src%int)) then
          eval%int_in2 => eval_target%int
       end if
    end if
    call interaction_reassign_links (eval%int, eval_src%int, eval_target%int)
  end subroutine evaluator_reassign_links_eval

  subroutine evaluator_reassign_links_int (eval, int_src, int_target)
    type(evaluator_t), intent(inout) :: eval
    type(interaction_t), intent(in) :: int_src
    type(interaction_t), intent(in), target :: int_target
    if (associated (eval%int_in1)) then
       if (interaction_get_tag (eval%int_in1) &
            == interaction_get_tag (int_src)) then
          eval%int_in1 => int_target
       end if
    end if
    if (associated (eval%int_in2)) then
       if (interaction_get_tag (eval%int_in2) &
            == interaction_get_tag (int_src)) then
          eval%int_in2 => int_target
       end if
    end if
    call interaction_reassign_links (eval%int, int_src, int_target)
  end subroutine evaluator_reassign_links_int

  subroutine evaluator_get_unstable_particle (eval, flv, p, i)
    type(evaluator_t), intent(in) :: eval
    type(flavor_t), intent(out) :: flv
    type(vector4_t), intent(out) :: p
    integer, intent(out) :: i
    call interaction_get_unstable_particle (eval%int, flv, p, i)
  end subroutine evaluator_get_unstable_particle

  elemental subroutine evaluator_final (eval)
    type(evaluator_t), intent(inout) :: eval
    call interaction_final (eval%int)
  end subroutine evaluator_final

  subroutine evaluator_evaluate (eval)
    type(evaluator_t), intent(inout), target :: eval
    integer :: i
    select case (eval%type)
    case (EVAL_PRODUCT)
       do i = 1, size(eval%pairing_array)
          call interaction_evaluate_product (eval%int, i, &
               eval%int_in1, eval%int_in2, &
               eval%pairing_array(i)%i1, eval%pairing_array(i)%i2)
       end do
    case (EVAL_SQUARE_WITH_COLOR_FACTORS)
       do i = 1, size(eval%pairing_array)
          call interaction_evaluate_product_cf (eval%int, i, &
               eval%int_in1, eval%int_in1, &
               eval%pairing_array(i)%i1, eval%pairing_array(i)%i2, &
               eval%pairing_array(i)%factor)
       end do
    case (EVAL_SQUARED_FLOWS)
       do i = 1, size(eval%pairing_array)
          call interaction_evaluate_square_c (eval%int, i, &
               eval%int_in1, &
               eval%pairing_array(i)%i1)
       end do
    end select
  end subroutine evaluator_evaluate

  subroutine evaluator_test (mdl)
    type(model_t), intent(in), target :: mdl
    call evaluator_test1 (mdl)
  end subroutine evaluator_test

  subroutine evaluator_test1 (mdl)
    type(model_t), intent(in), target :: mdl
    type(interaction_t), target :: int_qqtt, int_tbw
    type(flavor_t), dimension(:), allocatable :: flv
    type(color_t), dimension(:), allocatable :: col
    type(helicity_t), dimension(:), allocatable :: hel
    type(quantum_numbers_t), dimension(:), allocatable :: qn
    integer :: f, c, h1, h2, h3
!     type(vector4_t), dimension(4) :: p
!     type(vector4_t), dimension(2) :: q
    type(quantum_numbers_mask_t) :: qn_mask_conn
    type(evaluator_t), target :: eval
    print *, "*** Evaluator for matrix product"
    print *, "***   Construct interaction for qq -> tt"
    call interaction_init (int_qqtt, 2, 0, 2, set_relations=.true.)
    allocate (flv (4), col (4), hel (4), qn (4))
    do c = 1, 2
       select case (c)
       case (1)
          call color_init_col_acl (col, (/ 1, 0, 1, 0 /), (/ 0, 2, 0, 2 /))
       case (2)
          call color_init_col_acl (col, (/ 1, 0, 2, 0 /), (/ 0, 1, 0, 2 /))
       end select
       do f = 1, 2
          call flavor_init (flv, (/f, -f, 6, -6/), mdl)
          do h1 = -1, 1, 2
             call helicity_init (hel(3), h1)
             do h2 = -1, 1, 2
                call helicity_init (hel(4), h2)
                call quantum_numbers_init (qn, flv, col, hel)
                call interaction_add_state (int_qqtt, qn)
             end do
          end do
       end do
    end do
    call interaction_freeze (int_qqtt)
    deallocate (flv, col, hel, qn)
    print *, "***   Construct interaction for t -> bW"
    call interaction_init (int_tbw, 1, 0, 2, set_relations=.true.)
    allocate (flv (3), col (3), hel (3), qn (3))
    call flavor_init (flv, (/ 6, 5, 24 /), mdl)
    call color_init_col_acl (col, (/ 1, 1, 0 /), (/ 0, 0, 0 /))
    do h1 = -1, 1, 2
       call helicity_init (hel(1), h1)
       do h2 = -1, 1, 2
          call helicity_init (hel(2), h2)
          do h3 = -1, 1
             call helicity_init (hel(3), h3)
             call quantum_numbers_init (qn, flv, col, hel)
             call interaction_add_state (int_tbw, qn)
          end do
       end do
    end do
    call interaction_freeze (int_tbw)
    deallocate (flv, col, hel, qn)
    print *, "***   Link interactions"
    call interaction_set_source_link (int_tbw, 1, int_qqtt, 3)
    qn_mask_conn = new_quantum_numbers_mask (.false.,.false.,.true.)
    print *, "***   Show input"
    call interaction_write (int_qqtt)
    print *
    call interaction_write (int_tbw)
    print *
    print *, "***   Evaluate product"
    call evaluator_init_product &
         (eval, int_qqtt, int_tbw, qn_mask_conn)
    call evaluator_write (eval)

!     p(1) = vector4_moving (1000._default, 1000._default, 3)
!     p(2) = vector4_moving (200._default, 200._default, 2)
!     p(3) = vector4_moving (100._default, 200._default, 1)
!     p(4) = p(1) - p(2) - p(3)
!     call interaction_set_momenta (int1, p)
!     q(1) = vector4_moving (50._default,-50._default, 3)
!     q(2) = p(2) + p(4) - q(1)
!     call interaction_set_momenta (int2, q, outgoing=.true.)
!     call interaction_set_matrix_element &
!          (int1, (/(2._default,0._default), (4._default,1._default), (-3._default,0._default)/))
!     call interaction_set_matrix_element &
!          (int2, (/(-3._default,0._default), (0._default,1._default), (1._default,2._default)/))
!     call evaluator_receive_momenta (eval)
!     call evaluator_evaluate (eval)
!     call interaction_write (int1)
!     print *
!     call interaction_write (int2)
!     print *
!     call evaluator_write (eval)
!     print *
!     call interaction_final (int1)
!     call interaction_final (int2)
!     call evaluator_final (eval)

!     print *
!     print *, "*** Evaluator for matrix square"
!     call interaction_init (int1, 2, 0, 2, set_relations=.true.)
!     call flavor_init (flv, (/1, -1, 21, 21/), mdl)
!     call color_init (col(1), (/1/))
!     call color_init (col(2), (/-2/))
!     call color_init (col(3), (/2, -3/))
!     call color_init (col(4), (/3, -1/))
!     call quantum_numbers_init (qn, flv, col)
!     call interaction_add_state (int1, qn)
!     call color_init (col(3), (/3, -1/))
!     call color_init (col(4), (/2, -3/))
!     call quantum_numbers_init (qn, flv, col)
!     call interaction_add_state (int1, qn)
!     call color_init (col(3), (/2, -1/))
!     call color_init (col(4), .true.)
!     call quantum_numbers_init (qn, flv, col)
!     call interaction_add_state (int1, qn)
!     call interaction_freeze (int1)
!     ! qn_mask2 = all false (default)
!     call evaluator_init_square_with_color_factors (eval, int1, qn_mask2, nc=3)
!     call evaluator_init_squared_flows (eval2, int1, qn_mask2)
!     qn_mask2 = new_quantum_numbers_mask (.false., .true., .true.)
!     call evaluator_init_trace (eval3, eval%int, qn_mask2)
!     call interaction_set_matrix_element &
!          (int1, (/(2._default,0._default), (4._default,1._default), (-3._default,0._default)/))
!     call interaction_set_momenta (int1, p)
!     call interaction_write (int1)
!     print *
!     call evaluator_receive_momenta (eval)
!     call evaluator_evaluate (eval)
!     call evaluator_write (eval)
!     print *
!     call evaluator_receive_momenta (eval2)
!     call evaluator_evaluate (eval2)
!     call evaluator_write (eval2)
!     print *
!     call evaluator_receive_momenta (eval3)
!     call evaluator_evaluate (eval3)
!     call evaluator_write (eval3)
!     call interaction_final (int1)
!     call evaluator_final (eval)
!     call evaluator_final (eval2)
!     call evaluator_final (eval3)
  end subroutine evaluator_test1


end module evaluators
