! WHIZARD 2.0.0 Mon Aug 24 2009
! 
! (C) 1999-2009 by 
!     Wolfgang Kilian <kilian@hep.physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@physik.uni-freiburg.de>
!     with contributions by Sebastian Schmidt
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module processes

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils
  use diagnostics
  use os_interface
  use vamp_equivalences !NODEP!
  use vamp !NODEP!
  use lexers
  use parser
  use constants !NODEP!
  use lorentz
  use prt_lists
  use variables
  use expressions
  use models
  use flavors
  use quantum_numbers
  use polarizations
  use interactions
  use evaluators
  use particles
  use beams
  use sf_isr
  use sf_epa
  use sf_ewa
  use sf_lhapdf
  use strfun
  use mappings
  use phs_forests
  use cascades
  use process_libraries
  use hard_interactions

  implicit none
  private

  public :: integration_results_t
  public :: integration_results_append
  public :: process_t
  public :: process_write
  public :: process_is_valid
  public :: process_get_id
  public :: process_get_model_ptr
  public :: process_get_n_in
  public :: process_get_n_out
  public :: process_get_n_parameters
  public :: process_get_n_channels
  public :: process_get_sqme
  public :: process_get_n_calls
  public :: process_get_integral
  public :: process_get_error
  public :: process_get_accuracy
  public :: process_get_chi2
  public :: process_get_efficiency
  public :: process_get_eval_sqme_ptr
  public :: process_get_eval_flows_ptr
  public :: process_get_hi_int_ptr
  public :: process_get_hi_eval_sqme_ptr
  public :: process_get_hi_eval_flows_ptr
  public :: process_setup_beams
  public :: process_set_beam_momenta
  public :: process_set_strfun
  public :: process_set_strfun_mapping
  public :: process_connect_strfun
  public :: process_setup_phase_space
  public :: process_setup_cuts
  public :: process_setup_weight
  public :: process_setup_scale
  public :: process_setup_analysis
  public :: grid_parameters_t
  public :: process_setup_grids
  public :: process_integrate
  public :: process_setup_event_generation
  public :: process_generate_weighted_event
  public :: process_generate_unweighted_event
  public :: process_results_write_header
  public :: process_results_write_current
  public :: process_results_write_current_average
  public :: process_results_write_footer
  public :: process_results_write
  public :: process_record_integral
  public :: process_request_copy
  public :: process_copy_free
  public :: process_store_final
  public :: process_store_write
  public :: process_store_write_results
  public :: process_store_get_process_ptr
  public :: process_store_init_process
  public :: process_test

  integer, parameter :: PRC_UNKNOWN = 0
  integer, parameter :: PRC_DECAY = 1
  integer, parameter :: PRC_SCATTERING = 2

  integer, parameter :: RESULTS_CHUNK_SIZE = 10


  type :: integration_entry_t
     private
     integer :: process_type = PRC_UNKNOWN
     integer :: pass = 0
     integer :: it = 0
     integer :: n_it = 0
     integer :: n_calls = 0
     logical :: improved = .false.
     real(default) :: integral = 0
     real(default) :: error = 0
     real(default) :: efficiency = 0
     real(default) :: chi2 = 0
  end type integration_entry_t

  type :: integration_results_t
     private
     integer :: n_pass = 0
     integer :: n_it = 0
     type(integration_entry_t), dimension(:), allocatable :: entry
     type(integration_entry_t), dimension(:), allocatable :: average
  end type integration_results_t

  type :: process_t
     private
     integer :: type = PRC_UNKNOWN
     type(process_t), pointer :: copy => null ()
     logical :: is_original = .true.
     type(process_t), pointer :: original => null ()
     type(process_t), pointer :: working_copy => null ()
     logical :: in_use = .true.
     logical :: initialized = .false.
     logical :: use_beams = .true.
     logical :: beams_are_set = .false.
     type(string_t) :: id
     type(process_library_t), pointer :: prc_lib => null ()
     integer :: lib_index = 0
     integer :: store_index = 0
     type(model_t), pointer :: model
     integer :: n_strfun = 0
     integer :: n_par_strfun = 0
     integer :: n_par_hi = 0
     integer :: n_par = 0
     logical :: azimuthal_dependence = .false.
     logical :: vamp_grids_defined = .false.
     logical :: sqrts_known = .false.
     logical :: sqrts_hat_known = .false.
     real(default) :: sqrts = 0
     real(default) :: sqrts_hat = 0
     real(default), dimension(:), allocatable :: x_strfun
     real(default), dimension(:), allocatable :: x_hi
     integer :: n_channels = 0
     integer :: channel = 0
     logical :: lab_is_cm_frame = .true.
     type(lorentz_transformation_t) :: lt_cm_to_lab = identity
     real(default), dimension(:,:), allocatable :: x
     real(default), dimension(:), allocatable :: phs_factor
     real(default), dimension(:), allocatable :: mass_in
     real(default) :: flux_factor = 0
     real(default) :: averaging_factor = 0
     real(default) :: sf_mapping_factor = 0
     real(default) :: phs_volume = 0
     real(default) :: vamp_phs_factor = 0
     real(default) :: sqme = 0
     real(default) :: reweighting_factor = 0
     real(default) :: sample_function_value = 0
     real(default) :: scale = 0
     type(strfun_chain_t) :: sfchain
     type(hard_interaction_t) :: hi
     type(evaluator_t) :: eval_trace
     type(evaluator_t) :: eval_sqme
     type(evaluator_t) :: eval_flows
     type(phs_forest_t) :: forest
     type(vamp_equivalences_t) :: vamp_eq
     type(prt_list_t) :: prt_list
     type(var_list_t) :: var_list
     type(eval_tree_t) :: cut_expr
     type(eval_tree_t) :: reweighting_expr
     type(eval_tree_t) :: scale_expr
     type(eval_tree_t) :: analysis_expr
     logical, dimension(:), allocatable :: active_channel
     type(vamp_grids) :: grids
     type(integration_results_t) :: results
  end type process_t

  type :: grid_parameters_t
     integer :: min_calls_per_channel = 0
     integer :: min_calls_per_bin = 10
     integer :: min_bins = 3
     integer :: max_bins = 20
     logical :: stratified = .true.
     real(default) :: channel_weights_power = 0.25_default
  end type grid_parameters_t

  type :: process_entry_t
     type(process_t) :: process
     type(process_entry_t), pointer :: next => null ()
  end type process_entry_t

  type :: process_p
     type(process_t), pointer :: ptr => null ()
  end type process_p

  type :: process_store_t
     integer :: n = 0
     type(process_entry_t), pointer :: first => null ()
     type(process_entry_t), pointer :: last => null ()
     type(process_p), dimension(:), allocatable :: proc
  end type process_store_t


  type(process_store_t), save :: store


  interface process_set_strfun
     module procedure process_set_strfun_isr
     module procedure process_set_strfun_epa
     module procedure process_set_strfun_lhapdf
  end interface

  interface process_store_get_process_ptr
     module procedure process_store_get_process_ptr_int
     module procedure process_store_get_process_ptr_id
  end interface


contains

  subroutine integration_entry_init (entry, &
       process_type, pass, it, n_it, n_calls, improved, &
       integral, error, efficiency, chi2)
    type(integration_entry_t), intent(out) :: entry
    integer, intent(in) :: process_type, pass, it, n_it, n_calls
    logical, intent(in) :: improved
    real(default), intent(in) :: integral, error, efficiency
    real(default), intent(in), optional :: chi2
    entry%process_type = process_type
    entry%pass = pass
    entry%it = it
    entry%n_it = n_it
    entry%n_calls = n_calls
    entry%improved = improved
    entry%integral = integral
    entry%error = error
    entry%efficiency = efficiency
    if (present (chi2)) &
         entry%chi2 = chi2
  end subroutine integration_entry_init

  function integration_entry_get_n_calls (entry) result (n)
    integer :: n
    type(integration_entry_t), intent(in) :: entry
    n = entry%n_calls
  end function integration_entry_get_n_calls

  function integration_entry_get_integral (entry) result (int)
    real(default) :: int
    type(integration_entry_t), intent(in) :: entry
    int = entry%integral
  end function integration_entry_get_integral

  function integration_entry_get_error (entry) result (err)
    real(default) :: err
    type(integration_entry_t), intent(in) :: entry
    err = entry%error
  end function integration_entry_get_error

  function integration_entry_get_relative_error (entry) result (err)
    real(default) :: err
    type(integration_entry_t), intent(in) :: entry
    if (entry%integral /= 0) then
       err = entry%error / entry%integral
    else
       err = 0
    end if
  end function integration_entry_get_relative_error

  function integration_entry_get_accuracy (entry) result (acc)
    real(default) :: acc
    type(integration_entry_t), intent(in) :: entry
    acc = accuracy (entry%integral, entry%error, entry%n_calls)
  end function integration_entry_get_accuracy

  function accuracy (integral, error, n_calls) result (acc)
    real(default) :: acc
    real(default), intent(in) :: integral, error
    integer, intent(in) :: n_calls
    if (integral /= 0) then
       acc = error / integral * sqrt (real (n_calls, default))
    else
       acc = 0
    end if
  end function accuracy

  function integration_entry_get_efficiency (entry) result (eff)
    real(default) :: eff
    type(integration_entry_t), intent(in) :: entry
    eff = entry%efficiency
  end function integration_entry_get_efficiency

  function integration_entry_get_chi2 (entry) result (chi2)
    real(default) :: chi2
    type(integration_entry_t), intent(in) :: entry
    chi2 = entry%chi2
  end function integration_entry_get_chi2

  function integration_entry_has_improved (entry) result (flag)
    logical :: flag
    type(integration_entry_t), intent(in) :: entry
    flag = entry%improved
  end function integration_entry_has_improved

  subroutine write_header (process_type, unit, logfile)
    integer, intent(in) :: process_type
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: logfile
    character(5) :: phys_unit
    integer :: u
    u = output_unit (unit);  if (u < 0)  return
    select case (process_type)
    case (PRC_DECAY);      phys_unit = "[GeV]"
    case (PRC_SCATTERING); phys_unit = "[fb] "
    case default
       phys_unit = ""
    end select
    write (msg_buffer, "(A)") &
         "It      Calls  Integral" // phys_unit // &
         " Error" // phys_unit // &
         "  Err[%]    Acc  Eff[%]   Chi2 N[It] |"
    call msg_message (unit=u, logfile=logfile)
  end subroutine write_header
       
  subroutine write_hline (unit)
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(A)")  "|" // (repeat ("-", 77)) // "|"
  end subroutine write_hline
  
  subroutine write_dline (unit)
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(A)")  "|" // (repeat ("=", 77)) // "|"
  end subroutine write_dline
  
  subroutine integration_entry_write (entry, unit)
    type(integration_entry_t), intent(in) :: entry
    integer, intent(in), optional :: unit
    integer :: u
    character(1) :: star
    u = output_unit (unit);  if (u < 0)  return
    if (entry%improved) then
       star = "*"
    else
       star = " "
    end if
1   format (1x, I3, 1x, I10, 1x, 1PE14.7, 1x, 1PE9.2, 1x, 2PF7.2, &
            1x, 0PF7.2, A1, 1x, 2PF6.2, 1x, 0PF7.2, 1x, I3)
    if (entry%n_it /= 1) then
       write (u, 1) &
            entry%it, &
            entry%n_calls, &
            entry%integral, &
            entry%error, &
            integration_entry_get_relative_error (entry), &
            integration_entry_get_accuracy (entry), &
            star, &
            entry%efficiency, &
            entry%chi2, &
            entry%n_it
    else
       write (u, 1) &
            entry%it, &
            entry%n_calls, &
            entry%integral, &
            entry%error, &
            integration_entry_get_relative_error (entry), &
            integration_entry_get_accuracy (entry), &
            star, &
            entry%efficiency
    end if
  end subroutine integration_entry_write

  function compute_average (entry, pass) result (result)
    type(integration_entry_t) :: result
    type(integration_entry_t), dimension(:), intent(in) :: entry
    integer, intent(in) :: pass
    integer :: i
    logical, dimension(size(entry)) :: mask
    real(default), dimension(size(entry)) :: ivar
    real(default) :: sum_ivar, variance
    result%process_type = entry(1)%process_type
    mask = entry%pass == pass
    result%it = maxval (entry%it, mask)
    result%n_it = count (mask)
    result%n_calls = sum (entry%n_calls, mask)
    where (entry%error /= 0)
       ivar = 1 / entry%error ** 2
    elsewhere
       ivar = 0
    end where
    sum_ivar = sum (ivar, mask)
    if (sum_ivar /= 0) then
       variance = 1 / sum_ivar
    else
       variance = 0
    end if
    result%integral = sum (entry%integral * ivar, mask) * variance
    result%error = sqrt (variance)
    if (result%n_it > 1) then
       result%chi2 = sum ((entry%integral - result%integral)**2 * ivar, mask) &
                     / (result%n_it - 1)
    end if
    do i = size (entry), 1, -1
       if (mask(i)) then
          result%efficiency = entry(i)%efficiency
          exit
       end if
    end do
  end function compute_average

  subroutine integration_results_init (results)
    type(integration_results_t), intent(out) :: results
    results%n_pass = 0
    results%n_it = 0
    allocate (results%entry (RESULTS_CHUNK_SIZE))
    allocate (results%average (RESULTS_CHUNK_SIZE))
  end subroutine integration_results_init

  subroutine integration_results_write (results, unit)
    type(integration_results_t), intent(in) :: results
    integer, intent(in), optional :: unit
    integer :: u, n
    u = output_unit (unit);  if (u < 0)  return
    call write_dline (unit)
    if (results%n_it /= 0) then
       call write_header (results%entry(1)%pass, unit)
       call write_dline (unit)
       do n = 1, results%n_it
          if (n > 1) then
             if (results%entry(n)%pass /= results%entry(n-1)%pass) then
                call write_hline (unit)
                call integration_entry_write &
                     (results%average(results%entry(n-1)%pass), unit)
                call write_hline (unit)
             end if
          end if
          call integration_entry_write (results%entry(n), unit)
       end do
       call write_dline(unit)
       call integration_entry_write (results%average(results%n_pass), unit)
    else
       call msg_message ("[WHIZARD integration results: empty]", unit)
    end if
    call write_dline (unit)
  end subroutine integration_results_write

  subroutine integration_results_write_current (results, unit)
    type(integration_results_t), intent(in) :: results
    integer, intent(in), optional :: unit
    integer :: u, n
    u = output_unit (unit);  if (u < 0)  return
    n = results%n_it
    if (n /= 0)  call integration_entry_write (results%entry(n), unit)
  end subroutine integration_results_write_current

  subroutine integration_results_write_current_average (results, unit)
    type(integration_results_t), intent(in) :: results
    integer, intent(in), optional :: unit
    integer :: u, n
    u = output_unit (unit);  if (u < 0)  return
    n = results%n_pass
    if (n /= 0)  call integration_entry_write (results%average(n), unit)
  end subroutine integration_results_write_current_average

  subroutine integration_results_expand (results)
    type(integration_results_t), intent(inout) :: results
    type(integration_entry_t), dimension(:), allocatable :: entry_tmp
    if (results%n_it == size (results%entry)) then
       allocate (entry_tmp (results%n_it))
       entry_tmp = results%entry
       deallocate (results%entry)
       allocate (results%entry (results%n_it + RESULTS_CHUNK_SIZE))
       results%entry(:results%n_it) = entry_tmp
       deallocate (entry_tmp)
    end if
    if (results%n_pass == size (results%average)) then
       allocate (entry_tmp (results%n_pass))
       entry_tmp = results%average
       deallocate (results%average)
       allocate (results%average (results%n_it + RESULTS_CHUNK_SIZE))
       results%average(:results%n_pass) = entry_tmp
       deallocate (entry_tmp)
    end if
  end subroutine integration_results_expand

  subroutine integration_results_append_entry (results, entry)
    type(integration_results_t), intent(inout) :: results
    type(integration_entry_t), intent(in) :: entry
    if (results%n_it == 0) then
       call integration_results_init (results)
       results%n_it = 1
       results%n_pass = 1
    else
       call integration_results_expand (results)
       results%n_it = results%n_it + 1
       if (entry%pass /= results%entry(results%n_it)%pass) &
            results%n_pass = results%n_pass + 1
    end if
    results%entry(results%n_it) = entry
    results%average(results%n_pass) = &
         compute_average (results%entry, entry%pass)
  end subroutine integration_results_append_entry

  subroutine integration_results_append (results, &
       process_type, pass, n_it, n_calls, &
       integral, error, efficiency)
    type(integration_results_t), intent(inout) :: results
    integer, intent(in) :: process_type, pass, n_it, n_calls
    real(default), intent(in) :: integral, error, efficiency
    logical :: improved
    type(integration_entry_t) :: entry
    if (results%n_it /= 0) then
       improved = accuracy (integral, error, n_calls) &
            < integration_entry_get_accuracy (results%entry(results%n_it))
    else
       improved = .true.
    end if
    call integration_entry_init (entry, &
         process_type, pass, results%n_it+1, n_it, n_calls, improved, & 
         integral, error, efficiency)
    call integration_results_append_entry (results, entry)
  end subroutine integration_results_append
         
  function integration_results_get_n_calls (results) result (n_calls)
    integer :: n_calls
    type(integration_results_t), intent(in) :: results
    n_calls = integration_entry_get_n_calls (results%average(results%n_pass))
  end function integration_results_get_n_calls

  function integration_results_get_integral (results) result (integral)
    real(default) :: integral
    type(integration_results_t), intent(in) :: results
    integral = integration_entry_get_integral (results%average(results%n_pass))
  end function integration_results_get_integral

  function integration_results_get_error (results) result (error)
    real(default) :: error
    type(integration_results_t), intent(in) :: results
    error = integration_entry_get_error (results%average(results%n_pass))
  end function integration_results_get_error

  function integration_results_get_accuracy (results) result (accuracy)
    real(default) :: accuracy
    type(integration_results_t), intent(in) :: results
    accuracy = integration_entry_get_accuracy (results%average(results%n_pass))
  end function integration_results_get_accuracy

  function integration_results_get_chi2 (results) result (chi2)
    real(default) :: chi2
    type(integration_results_t), intent(in) :: results
    chi2 = integration_entry_get_chi2 (results%average(results%n_pass))
  end function integration_results_get_chi2

  function integration_results_get_efficiency (results) result (efficiency)
    real(default) :: efficiency
    type(integration_results_t), intent(in) :: results
    efficiency = &
         integration_entry_get_efficiency (results%average(results%n_pass))
  end function integration_results_get_efficiency

  subroutine process_init &
       (process, prc_lib, process_lib_index, process_store_index, &
        process_id, model, var_list, use_beams)
    type(process_t), intent(out), target :: process
    type(process_library_t), intent(in), target :: prc_lib
    integer, intent(in) :: process_lib_index
    integer, intent(in) :: process_store_index
    type(string_t), intent(in) :: process_id
    type(model_t), intent(in), target :: model
    type(var_list_t), intent(in), optional, target :: var_list
    logical, intent(in), optional :: use_beams
    integer :: n_in, n_out, n_tot
    type(flavor_t), dimension(:), allocatable :: flv_in
    process%prc_lib => prc_lib
    process%lib_index = process_lib_index
    process%store_index = process_store_index
    process%id = process_id
    call hard_interaction_init &
         (process%hi, prc_lib, process_lib_index, process_id, model)
    process%model => hard_interaction_get_model_ptr (process%hi)
    if (.not. hard_interaction_is_valid (process%hi))  return
    process%id = hard_interaction_get_id (process%hi)
    if (present (use_beams))  process%use_beams = use_beams
    n_in  = hard_interaction_get_n_in  (process%hi)
    n_out = hard_interaction_get_n_out (process%hi)
    n_tot = hard_interaction_get_n_tot (process%hi)
    select case (n_in)
    case (1);  process%type = PRC_DECAY
    case (2);  process%type = PRC_SCATTERING
    end select
    allocate (flv_in (n_in), process%mass_in (n_in))
    call flavor_init (flv_in, hard_interaction_get_first_pdg_in (process%hi), &
                      process%model)
    process%mass_in = flavor_get_mass (flv_in)
    if (process%use_beams) then
       process%averaging_factor = 1
    else
       process%averaging_factor = &
            1._default / product (flavor_get_multiplicity (flv_in))
    end if
    if (present (var_list)) then
       call var_list_link (process%var_list, var_list)
    else
       call var_list_link (process%var_list, &
            model_get_var_list_ptr (process%model))
    end if
    call var_list_append_int (process%var_list, var_str ("n_in"),  n_in)
    call var_list_append_int (process%var_list, var_str ("n_out"), n_out)
    call var_list_append_int (process%var_list, var_str ("n_tot"), n_tot)
    call var_list_append_real_ptr (process%var_list, &
         var_str ("sqrts"), process%sqrts, process%sqrts_known)
    call var_list_append_real_ptr (process%var_list, &
         var_str ("sqrts_hat"), process%sqrts_hat, process%sqrts_hat_known)
    call interaction_init_prt_list &
         (hard_interaction_get_int_ptr (process%hi), process%prt_list)
    call integration_results_init (process%results)
    process%initialized = .true.
  end subroutine process_init

  recursive subroutine process_final (process)
    type(process_t), intent(inout), target :: process
    if (associated (process%copy)) then
       call process_final (process%copy)
       deallocate (process%copy)
    end if
    process%initialized = .false.
    process%type = PRC_UNKNOWN
    process%sqrts_known = .false.
    process%sqrts_hat_known = .false.
    call strfun_chain_final (process%sfchain)
    call hard_interaction_final (process%hi)
    call evaluator_final (process%eval_trace)
    if (process%is_original) then
       call evaluator_final (process%eval_sqme)
       call evaluator_final (process%eval_flows)
    end if
    call phs_forest_final (process%forest)
    call vamp_equivalences_final (process%vamp_eq)
    if (process%is_original) then
       call var_list_final (process%var_list)
       call eval_tree_final (process%cut_expr)
       call eval_tree_final (process%reweighting_expr)
       call eval_tree_final (process%scale_expr)
       call eval_tree_final (process%analysis_expr)
    end if
    if (process%vamp_grids_defined) then
       call vamp_delete_grids (process%grids)
    end if
  end subroutine process_final

  subroutine process_write &
       (process, unit, verbose, show_momentum_sum, show_mass)
    type(process_t), intent(in) :: process
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: verbose, show_momentum_sum, show_mass
    integer :: u, i
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(A)")  repeat ("=", 72)
    write (u, *)  "Process data:", process%lib_index, &
         "(", char (process%id), ")"
    select case (process%type)
    case (PRC_UNKNOWN);     write (u, *) "  [unknown]"
    case (PRC_DECAY);       write (u, *) "  [decay]"
    case (PRC_SCATTERING);  write (u, *) "  [scattering]"
    end select
    write (u, *)  "  use separate beam setup = ", process%use_beams
    if (process%use_beams) then
       write (u, *)  "  number of structure functions  = ", process%n_strfun
       write (u, *)  "  number of strfun parameters    = ", process%n_par_strfun
    end if
    write (u, *)  "  number of process parameters   = ", process%n_par_hi
    write (u, *)  "  number of parameters total     = ", process%n_par
    write (u, *)  "  number of integration channels = ", process%n_channels
    if (process%sqrts_known) then
       write (u, *)  "  c.m. energy (sqrts)     = ", process%sqrts
    else
       write (u, *)  "  c.m. energy (sqrts)     = [unknown]"
    end if
    if (process%sqrts_hat_known) then
       write (u, *)  "  c.m. energy (sqrts_hat) = ", process%sqrts_hat
    else
       write (u, *)  "  c.m. energy (sqrts_hat) = [unknown]"
    end if
    if (allocated (process%mass_in)) then
       write (u, *)  "  Incoming parton masses  = ", process%mass_in
    else
       write (u, *)  "  Incoming parton masses  = [unknown]"
    end if
    write (u, *)  "  In-state flux factor    = ", process%flux_factor
    write (u, *)  "  Strfun mapping factor   = ", process%sf_mapping_factor
    if (.not. process%use_beams) then
       write (u, *)  "  Spin averaging factor   = ", process%averaging_factor
    end if
    write (u, *)  "  VAMP phs factor         = ", process%vamp_phs_factor
    write (u, *)  "  Phase-space volume      = ", process%phs_volume
    write (u, *)  "  Squared matrix element  = ", process%sqme
    write (u, *)  "  Reweighting factor      = ", process%reweighting_factor
    write (u, *)  "  Sample-function value   = ", &
         process%sample_function_value
    write (u, *)  repeat ("-", 72)
    if (process%use_beams) then
       write (u, *)  "Structure function parameters ="
       if (allocated (process%x_strfun)) then
          write (u, *)  process%x_strfun
       else
          write (u, *)  "[empty]"
       end if
    end if
    write (u, *)  "  Process energy scale    = ", process%scale
    write (u, *)  "Phase-space integration parameters (input) ="
    if (allocated (process%x_hi)) then
       write (u, *)  process%x_hi
    else
       write (u, *)  "[empty]"
    end if
    write (u, *)  "Integration channel =", process%channel
    write (u, *)  "Phase-space integration parameters (complete) ="
    if (allocated (process%x)) then
       do i = 1, size (process%x, 2)
          write (u, *)  process%x(:,i)
       end do
    else
       write (u, *)  "[empty]"
    end if
    if (.not. process%lab_is_cm_frame) then
       write (u, *)  "Tranformation c.m. -> lab ="
       call lorentz_transformation_write (process%lt_cm_to_lab, u)
    end if
    write (u, *)  "Channels: phase-space factors ="
    if (allocated (process%phs_factor)) then
       write (u, *)  process%phs_factor
    else
       write (u, *) "[not allocated]"
    end if
    write (u, "(A)")  repeat ("-", 72)
    if (process%use_beams) then
       call strfun_chain_write &
            (process%sfchain, unit, verbose, show_momentum_sum, show_mass)
       write (u, "(A)")  repeat ("-", 72)
    end if
    call hard_interaction_write &
         (process%hi, unit, verbose, show_momentum_sum, show_mass)
    write (u, "(A)")  repeat ("-", 72)
    if (process%use_beams) then
       write (u, "(A)") "Trace including color factors (beams + strfun + hard interaction)"
       call evaluator_write &
            (process%eval_trace, unit, verbose, show_momentum_sum, show_mass)
       write (u, "(A)")  repeat ("-", 72)
       write (u, "(A)") "Exclusive sqme including color factors (beams + strfun + hard interaction)"
       call evaluator_write &
            (process%eval_sqme, unit, verbose, show_momentum_sum, show_mass)
       write (u, "(A)")  repeat ("-", 72)
       write (u, "(A)") "Color flow coefficients (beams + strfun + hard interaction)"
       call evaluator_write &
            (process%eval_flows, unit, verbose, show_momentum_sum, show_mass)
       write (u, "(A)")  repeat ("-", 72)
    end if
    call phs_forest_write (process%forest, unit)
    write (u, "(A)")  repeat ("-", 72)
    call vamp_equivalences_write (process%vamp_eq, unit)
    write (u, "(A)")  repeat ("-", 72)
    call prt_list_write (process%prt_list, unit)
    write (u, "(A)")  repeat ("-", 72)
    call var_list_write (process%var_list, unit)
    write (u, "(A)")  repeat ("-", 72)
    write (u, "(A)")  "Cut expression:"
    call eval_tree_write (process%cut_expr, unit)
    write (u, "(A)")  repeat ("-", 72)
    write (u, "(A)")  "Weight expression:"
    call eval_tree_write (process%reweighting_expr, unit)
    write (u, "(A)")  repeat ("-", 72)
    write (u, "(A)")  "Scale expression:"
    call eval_tree_write (process%scale_expr, unit)
    write (u, "(A)")  repeat ("-", 72)
    write (u, "(A)")  "Analysis expression:"
    call eval_tree_write (process%analysis_expr, unit)
    write (u, "(A)")  repeat ("-", 72)
    if (process%vamp_grids_defined) then
       call vamp_write_grids (process%grids, u)
    else
       write (u, "(A)")  "VAMP grids: [empty]"
    end if
    write (u, *)
    call integration_results_write (process%results, unit)
  end subroutine process_write

  function process_is_valid (process) result (flag)
    logical :: flag
    type(process_t), intent(in) :: process
    flag = process%initialized
  end function process_is_valid

  function process_get_id (process) result (process_id)
    type(string_t) :: process_id
    type(process_t), intent(in) :: process
    process_id = process%id
  end function process_get_id

  function process_get_model_ptr (process) result (model)
    type(model_t), pointer :: model
    type(process_t), intent(in) :: process
    model => process%model
  end function process_get_model_ptr

  function process_get_n_in (process) result (n)
    integer :: n
    type(process_t), intent(in) :: process
    n = hard_interaction_get_n_in (process%hi)
  end function process_get_n_in

  function process_get_n_out (process) result (n)
    integer :: n
    type(process_t), intent(in) :: process
    n = hard_interaction_get_n_out (process%hi)
  end function process_get_n_out

  function process_get_n_parameters (process) result (n)
    integer :: n
    type(process_t), intent(in) :: process
    n = process%n_par
  end function process_get_n_parameters

  function process_get_n_channels (process) result (n)
    integer :: n
    type(process_t), intent(in) :: process
    n = process%n_channels
  end function process_get_n_channels

  function process_get_sqme (process) result (sqme)
    real(default) :: sqme
    type(process_t), intent(in) :: process
    sqme = process%sqme
  end function process_get_sqme

  function process_get_n_calls (process) result (n_calls)
    integer :: n_calls
    type(process_t), intent(in) :: process
    n_calls = integration_results_get_n_calls (process%results)
  end function process_get_n_calls

  function process_get_integral (process) result (integral)
    real(default) :: integral
    type(process_t), intent(in) :: process
    integral = integration_results_get_integral (process%results)
  end function process_get_integral

  function process_get_error (process) result (error)
    real(default) :: error
    type(process_t), intent(in) :: process
    error = integration_results_get_error (process%results)
  end function process_get_error

  function process_get_accuracy (process) result (accuracy)
    real(default) :: accuracy
    type(process_t), intent(in) :: process
    accuracy = integration_results_get_accuracy (process%results)
  end function process_get_accuracy

  function process_get_chi2 (process) result (chi2)
    real(default) :: chi2
    type(process_t), intent(in) :: process
    chi2 = integration_results_get_chi2 (process%results)
  end function process_get_chi2

  function process_get_efficiency (process) result (efficiency)
    real(default) :: efficiency
    type(process_t), intent(in) :: process
    efficiency = integration_results_get_efficiency (process%results)
  end function process_get_efficiency

  function process_get_eval_sqme_ptr (process) result (eval)
    type(evaluator_t), pointer :: eval
    type(process_t), intent(in), target :: process
    if (process%use_beams) then
       eval => process%eval_sqme
    else
       eval => hard_interaction_get_eval_sqme_ptr (process%hi)
    end if
  end function process_get_eval_sqme_ptr

  function process_get_eval_flows_ptr (process) result (eval)
    type(evaluator_t), pointer :: eval
    type(process_t), intent(in), target :: process
    if (process%use_beams) then
       eval => process%eval_flows
    else
       eval => hard_interaction_get_eval_flows_ptr (process%hi)
    end if
  end function process_get_eval_flows_ptr

  function process_get_hi_int_ptr (process) result (int)
    type(interaction_t), pointer :: int
    type(process_t), intent(in), target :: process
    int => hard_interaction_get_int_ptr (process%hi)
  end function process_get_hi_int_ptr

  function process_get_hi_eval_sqme_ptr (process) result (eval)
    type(evaluator_t), pointer :: eval
    type(process_t), intent(in), target :: process
    eval => hard_interaction_get_eval_sqme_ptr (process%hi)
  end function process_get_hi_eval_sqme_ptr

  function process_get_hi_eval_flows_ptr (process) result (eval)
    type(evaluator_t), pointer :: eval
    type(process_t), intent(in), target :: process
    eval => hard_interaction_get_eval_flows_ptr (process%hi)
  end function process_get_hi_eval_flows_ptr

  subroutine process_setup_beams &
       (process, beam_data, n_strfun, n_mapping, sqrts)
    type(process_t), intent(inout), target :: process
    type(beam_data_t), intent(in) :: beam_data
    integer, intent(in) :: n_strfun, n_mapping
    real(default), intent(in), optional :: sqrts
    if (process%use_beams) then
       process%sqrts = beam_data%sqrts
       process%sqrts_known = .true.
       process%n_strfun = n_strfun
       process%azimuthal_dependence = &
            .not. all (polarization_is_diagonal (beam_data%pol))
       process%lab_is_cm_frame = beam_data%lab_is_cm_frame .and. n_strfun == 0
       call strfun_chain_init (process%sfchain, beam_data, n_strfun, n_mapping)
    else
       select case (process%type)
       case (PRC_DECAY)
          process%sqrts = process%mass_in(1)
       case (PRC_SCATTERING)
          if (present (sqrts)) then
             process%sqrts = sqrts
          else
             call msg_fatal ("Process setup: neither beams nor sqrts are known")
          end if
       end select
       process%sqrts_known = .true.
    end if
  end subroutine process_setup_beams

  subroutine process_set_beam_momenta (process, p)
    type(process_t), intent(inout), target :: process
    type(vector4_t), dimension(:), intent(in) :: p
    type(interaction_t), pointer :: hi_int
    if (process%use_beams) then
       call strfun_chain_set_beam_momenta (process%sfchain, p)
    else
       hi_int => hard_interaction_get_int_ptr (process%hi)
       call interaction_set_momenta (hi_int, p, outgoing=.false.)
    end if
    process%sqrts_hat = process%sqrts
    process%lab_is_cm_frame = .false.
    process%beams_are_set = .true.
  end subroutine process_set_beam_momenta

  subroutine process_set_strfun_isr &
       (process, i, line, isr_data, n_parameters)
    type(process_t), intent(inout), target :: process
    integer, intent(in) :: i, line, n_parameters
    type(isr_data_t), intent(in) :: isr_data
    if (process%use_beams) then
       call strfun_chain_set_strfun &
            (process%sfchain, i, line, isr_data, n_parameters)
    end if
  end subroutine process_set_strfun_isr

  subroutine process_set_strfun_epa &
       (process, i, line, epa_data, n_parameters)
    type(process_t), intent(inout), target :: process
    integer, intent(in) :: i, line, n_parameters
!    type(epa_data_t), dimension(:), intent(in) :: epa_data
    type(epa_data_t), intent(in) :: epa_data
    if (process%use_beams) then
       call strfun_chain_set_strfun &
            (process%sfchain, i, line, (/ epa_data /), n_parameters)
    end if
  end subroutine process_set_strfun_epa

  subroutine process_set_strfun_lhapdf &
       (process, i, line, lhapdf_data, n_parameters)
    type(process_t), intent(inout), target :: process
    integer, intent(in) :: i, line, n_parameters
    type(lhapdf_data_t), intent(in) :: lhapdf_data
    if (process%use_beams) then
       call strfun_chain_set_strfun &
            (process%sfchain, i, line, lhapdf_data, n_parameters)
    end if
  end subroutine process_set_strfun_lhapdf

  subroutine process_set_strfun_mapping (process, i, index, type, par)
    type(process_t), intent(inout) :: process
    integer, intent(in) :: i
    integer, intent(in) :: type
    integer, dimension(:), intent(in) :: index
    real(default), dimension(:), intent(in) :: par
    if (process%use_beams) then
       call strfun_chain_set_mapping (process%sfchain, i, index, type, par)
    end if
  end subroutine process_set_strfun_mapping

  subroutine process_connect_strfun (process, ok)
    type(process_t), intent(inout), target :: process
    logical, intent(out), optional :: ok
    integer, dimension(:), allocatable :: coll_index
    type(quantum_numbers_mask_t), dimension(:), allocatable :: mask_in
    type(quantum_numbers_mask_t) :: mask_tr
    type(evaluator_t), pointer :: eval_sfchain, eval_hi
    type(interaction_t), pointer :: int_beam, int_hi
    integer :: n_in, i
    n_in  = hard_interaction_get_n_in  (process%hi)
    allocate (mask_in (n_in))
    if (process%use_beams) then
       call strfun_chain_make_evaluators (process%sfchain, ok)
       allocate (coll_index (n_in))
       coll_index = strfun_chain_get_colliding_particles (process%sfchain)
       mask_in = strfun_chain_get_colliding_particles_mask (process%sfchain)
    else
       mask_in = new_quantum_numbers_mask (.true., .true., .true.)
    end if
    call hard_interaction_init_trace (process%hi, mask_in)
    if (process%use_beams) then
       int_beam => strfun_chain_get_beam_int_ptr (process%sfchain)
       eval_sfchain => strfun_chain_get_last_evaluator_ptr (process%sfchain)
       eval_hi => hard_interaction_get_eval_trace_ptr (process%hi)
       int_hi => hard_interaction_get_int_ptr (process%hi)
       mask_tr = new_quantum_numbers_mask (.true., .true., .true.)
       if (associated (eval_sfchain)) then
          do i = 1, n_in
             call interaction_set_source_link &
                  (int_hi, i, eval_sfchain, coll_index(i))
             call evaluator_set_source_link &
                  (eval_hi, i, eval_sfchain, coll_index(i))
          end do
          call evaluator_init_product &
               (process%eval_trace, eval_sfchain, eval_hi, mask_tr, mask_tr)
          if (evaluator_is_empty (process%eval_trace)) then
             call msg_fatal ("Mismatch between structure functions and hard process")
             if (present (ok))  ok = .false.
             return
          end if
       else
          do i = 1, n_in
             call interaction_set_source_link &
                  (int_hi, i, int_beam, coll_index(i))
             call evaluator_set_source_link &
                  (eval_hi, i, int_beam, coll_index(i))
          end do
          call evaluator_init_product &
               (process%eval_trace, int_beam, eval_hi, mask_tr, mask_tr)
          if (evaluator_is_empty (process%eval_trace)) then
             call msg_fatal ("Mismatch between beams and hard process")
             if (present (ok))  ok = .false.
             return
          end if
       end if
       process%n_par_strfun = &
            strfun_chain_get_n_parameters_tot (process%sfchain)
       allocate (process%x_strfun (process%n_par_strfun))
    end if
    process%n_par = process%n_par_strfun + process%n_par_hi
    if (present (ok))  ok = .true.
  end subroutine process_connect_strfun

  subroutine process_setup_phase_space (process, &
       phs_par, mapping_defaults, filename_out, filename_in, ok)
    type(process_t), intent(inout), target :: process
    type(phs_parameters_t), intent(in) :: phs_par
    type(mapping_defaults_t), intent(in) :: mapping_defaults
    type(string_t), intent(in), optional :: filename_out, filename_in
    logical, intent(out), optional :: ok
    type(cascade_set_t) :: cascade_set
    logical :: variable_limits
    integer :: n_in, n_out, n_tot, n_flv
    type(flavor_t), dimension(:,:), allocatable :: flv
    integer :: n_par_strfun
    logical, dimension(:), allocatable :: strfun_rigid
    integer :: unit
    logical :: phs_ok
    phs_ok = .false.
    variable_limits = process%n_strfun /= 0
    n_in  = hard_interaction_get_n_in  (process%hi)
    n_out = hard_interaction_get_n_out (process%hi)
    n_tot = hard_interaction_get_n_tot (process%hi)
    n_flv = hard_interaction_get_n_flv (process%hi)
    allocate (flv (n_tot, n_flv))
    call flavor_init (flv, &
         hard_interaction_get_flv_states (process%hi), process%model)
    if (present (filename_in)) then
       unit = free_unit ()
       open (unit, file=char(filename_in), action="read", status="old")
       call phs_forest_read (process%forest, unit, &
            process%id, n_in, n_out, process%model, phs_ok)
       close (unit)
       if (.not. phs_ok) then
          call msg_fatal ("Phase space file '" // char (filename_in) &
               // "': No valid phase space for process '" &
               // char (process%id) // "'")
          if (present (ok))  ok = .false.
          return
       end if
    end if
    if (.not. phs_ok) then
       unit = free_unit ()
       if (present (filename_out)) then
          open (unit, file=char(filename_out), &
                action="readwrite", status="replace")
       else
          open (unit, action="readwrite", status="scratch")
       end if
       call cascade_set_generate (cascade_set, &
            process%model, n_in, n_out, flv, process%sqrts, phs_par)
       if (.not. cascade_set_is_valid (cascade_set)) then
          call msg_error ("Process '" // char (process%id) &
               // "': no valid phase-space channels found")
          if (present (ok))  ok = .false.
          return
       end if
       write (unit, *) "process ", char (process%id)
       call cascade_set_write_file_format (cascade_set, unit)
       rewind (unit)
       call phs_forest_read (process%forest, unit, &
            process%id, n_in, n_out, process%model, phs_ok)
       close (unit)
       if (.not. phs_ok) then
          call msg_bug ("Generated phase space file: " &
               // "No valid phase space for process '" &
               // char (process%id) // "'")
       end if
    end if
    call phs_forest_set_flavors (process%forest, flv(:,1))
    call phs_forest_set_parameters &
         (process%forest, mapping_defaults, variable_limits)
    call phs_forest_set_equivalences (process%forest)
    if (process%use_beams) then
       n_par_strfun = strfun_chain_get_n_parameters_tot (process%sfchain)
       allocate (strfun_rigid (n_par_strfun))
       strfun_rigid = strfun_chain_dimension_is_rigid (process%sfchain)
    else
       n_par_strfun = 0
       allocate (strfun_rigid (0))
    end if
    call phs_forest_setup_vamp_equivalences (process%forest, &
         n_par_strfun, strfun_rigid, &
         process%azimuthal_dependence, &
         process%vamp_eq)
    process%n_channels = phs_forest_get_n_channels (process%forest)
    process%n_par_hi = phs_forest_get_n_parameters (process%forest)
    process%n_par = process%n_par_strfun + process%n_par_hi
    allocate (process%x_hi (process%n_par_hi))
    allocate (process%x (process%n_par, process%n_channels))
    allocate (process%phs_factor (process%n_channels))
    allocate (process%active_channel (process%n_channels))
    process%active_channel = .true.
    if (present (ok))  ok = .true.
  end subroutine process_setup_phase_space

  subroutine process_setup_cuts (process, parse_node)
    type(process_t), intent(inout), target :: process
    type(parse_node_t), intent(in), target :: parse_node
    call eval_tree_init_lexpr &
         (process%cut_expr, parse_node, process%var_list, process%prt_list)
  end subroutine process_setup_cuts

  subroutine process_setup_weight (process, parse_node)
    type(process_t), intent(inout), target :: process
    type(parse_node_t), intent(in), target :: parse_node
    call eval_tree_init_expr &
         (process%reweighting_expr, parse_node, process%var_list, &
          process%prt_list)
  end subroutine process_setup_weight

  subroutine process_setup_scale (process, parse_node)
    type(process_t), intent(inout), target :: process
    type(parse_node_t), intent(in), target :: parse_node
    call eval_tree_init_expr &
         (process%scale_expr, parse_node, process%var_list, process%prt_list)
  end subroutine process_setup_scale

  subroutine process_setup_analysis (process, parse_node)
    type(process_t), intent(inout), target :: process
    type(parse_node_t), intent(in), target :: parse_node
    call eval_tree_init_lexpr &
         (process%analysis_expr, parse_node, process%var_list, process%prt_list)
  end subroutine process_setup_analysis

  subroutine process_setup_grids (process, grid_parameters, calls)
    type(process_t), intent(inout), target :: process
    type(grid_parameters_t), intent(in) :: grid_parameters
    integer, intent(in) :: calls
    integer, dimension(:), allocatable :: num_div
    real(default), dimension(:), allocatable :: weights
    real(default), dimension(:,:), allocatable :: region
    integer :: min_calls
    allocate (num_div (process%n_par))
    min_calls = grid_parameters%min_calls_per_bin * process%n_channels
    if (min_calls /= 0) then
       num_div =  max (grid_parameters%min_bins, &
            min (calls / min_calls, grid_parameters%max_bins))
    else
       num_div = grid_parameters%max_bins
    end if
    allocate (region (2, process%n_par))
    region(1,:) = 0
    region(2,:) = 1
    allocate (weights (process%n_channels))
    weights = 1
    call vamp_create_grids (process%grids, region, calls, weights, &
         num_div=num_div, stratified=grid_parameters%stratified)
    process%vamp_grids_defined = .true.
  end subroutine process_setup_grids

  subroutine process_set_kinematics (process, x_in, channel, ok)
    type(process_t), intent(inout), target :: process
    real(default), dimension(:), intent(in) :: x_in
    integer, intent(in) :: channel
    logical, intent(out) :: ok
    type(interaction_t), pointer :: int
    type(evaluator_t), pointer :: eval
    integer :: i
    process%x_hi = x_in(:process%n_par_hi)
    process%x_strfun = x_in(process%n_par_hi+1:)
!     process%x_strfun = x_in(:process%n_par_strfun)
!     process%x_hi = x_in(process%n_par_strfun+1:)
    process%channel = channel
    int => hard_interaction_get_int_ptr (process%hi)
    eval => hard_interaction_get_eval_trace_ptr (process%hi)
    if (process%use_beams) then
       call strfun_chain_set_kinematics (process%sfchain, process%x_strfun)
       call interaction_receive_momenta (int)
       process%beams_are_set = .true.
       process%sqrts_hat = sqrt (max (interaction_get_s (int), 0._default))
    else if (.not. process%beams_are_set) then
       select case (process%type)
       case (PRC_DECAY)
          call interaction_set_momenta (int, &
               (/ vector4_at_rest (process%mass_in(1)) /), &
               outgoing=.false.)
       case (PRC_SCATTERING)
          call interaction_set_momenta (int, &
               colliding_momenta (process%sqrts, process%mass_in), &
               outgoing=.false.)
       end select
       process%sqrts_hat = process%sqrts
    end if
    select case (process%type)
    case (PRC_DECAY)
       process%flux_factor = &
            twopi4 / (2 * process%mass_in(1))
    case (PRC_SCATTERING)
       process%flux_factor = &
            conv * twopi4 &
            / (2 * sqrt (lambda (process%sqrts_hat ** 2, &
                                 process%mass_in(1) ** 2, &
                                 process%mass_in(2) ** 2)))
    end select
    process%sqrts_hat_known = .true.
    if (.not. process%lab_is_cm_frame) then
       process%lt_cm_to_lab = interaction_get_cm_transformation (int)
       call phs_forest_set_prt_in (process%forest, int, process%lt_cm_to_lab)
    else
       call phs_forest_set_prt_in (process%forest, int)
    end if
    process%x(:process%n_par_hi,channel) = process%x_hi
    forall (i = 1 : process%n_par_strfun)
       process%x(process%n_par_hi+i,:) = process%x_strfun
    end forall
    call phs_forest_evaluate_phase_space (process%forest, &
         channel, process%active_channel, process%sqrts_hat, &
         process%x, process%phs_factor, process%phs_volume, ok)
    if (ok) then
       if (process%lab_is_cm_frame) then
          call phs_forest_get_prt_out (process%forest, int)
       else
          call phs_forest_get_prt_out &
               (process%forest, int, process%lt_cm_to_lab)
       end if
       call interaction_momenta_to_prt_list (int, process%prt_list)
       call process_compute_scale (process)
    end if
    call evaluator_receive_momenta (eval)
    if (process%use_beams) &
         call evaluator_receive_momenta (process%eval_trace)
  end subroutine process_set_kinematics

  subroutine process_compute_vamp_phs_factor (process, weights)
    type(process_t), intent(inout), target :: process
    real(default), dimension(:), intent(in) :: weights
    real(default), dimension(process%n_channels) :: vamp_prob
    real(default) :: dp
    integer :: i
    do i = 1, process%n_channels
       if (process%active_channel(i)) then
          vamp_prob(i) = &
               vamp_probability (process%grids%grids(i), process%x(:,i))
       else
          vamp_prob(i) = 0
       end if
    end do
    dp = dot_product (weights, vamp_prob / process%phs_factor)
    if (dp /= 0) then
       process%vamp_phs_factor = vamp_prob(process%channel) / dp
    else
       process%vamp_phs_factor = 0
    end if
  end subroutine process_compute_vamp_phs_factor

  function process_passes_cuts (process) result (flag)
    logical :: flag
    type(process_t), intent(inout), target :: process
    if (eval_tree_is_defined (process%cut_expr)) then
       call eval_tree_evaluate (process%cut_expr)
       if (eval_tree_result_is_known (process%cut_expr)) then
          flag = eval_tree_get_log (process%cut_expr)
       else
          flag = .true.
       end if
    else
       flag = .true.
    end if
  end function process_passes_cuts

  subroutine process_compute_reweighting_factor (process)
    type(process_t), intent(inout), target :: process
    if (eval_tree_is_defined (process%reweighting_expr)) then
       call eval_tree_evaluate (process%reweighting_expr)
       if (eval_tree_result_is_known (process%reweighting_expr)) then
          process%reweighting_factor = &
               eval_tree_get_real (process%reweighting_expr)
       else
          process%reweighting_factor = 1
       end if
    else
       process%reweighting_factor = 1
    end if
  end subroutine process_compute_reweighting_factor

  subroutine process_do_analysis (process) 
    type(process_t), intent(inout), target :: process
    if (eval_tree_is_defined (process%analysis_expr)) then
       call eval_tree_evaluate (process%analysis_expr)
    end if
  end subroutine process_do_analysis

  subroutine process_compute_scale (process)
    type(process_t), intent(inout), target :: process
    if (eval_tree_is_defined (process%scale_expr)) then
       call eval_tree_evaluate (process%scale_expr)
       if (eval_tree_result_is_known (process%scale_expr)) then
          process%scale = eval_tree_get_real (process%scale_expr)
       else
          process%scale = process%sqrts_hat
       end if
    else
       process%scale = process%sqrts_hat
    end if
  end subroutine process_compute_scale

  subroutine process_evaluate (process)
    type(process_t), intent(inout), target :: process
    if (process%use_beams) then
       call strfun_chain_evaluate (process%sfchain, process%scale)
       process%sf_mapping_factor = &
            strfun_chain_get_mapping_factor (process%sfchain)
    else
       process%sf_mapping_factor = 1
    end if
    call hard_interaction_evaluate (process%hi)
    if (process%use_beams) then
       call evaluator_evaluate (process%eval_trace)
       process%sqme = evaluator_sum (process%eval_trace)
    else
       process%sqme = evaluator_sum &
            (hard_interaction_get_eval_trace_ptr (process%hi)) &
            * process%averaging_factor
    end if
!    call process_write (process, 66); stop
  end subroutine process_evaluate

  function process_compute_sqme_sum (process, p) result (sqme)
    real(default) :: sqme
    type(process_t), intent(inout), target :: process
    type(vector4_t), dimension(:), intent(in) :: p
    sqme = hard_interaction_compute_sqme_sum (process%hi, p)
  end function process_compute_sqme_sum

  function process_get_vamp_efficiency_array (process) result (efficiency)
    real(default), dimension(:), allocatable :: efficiency
    type(process_t), intent(in) :: process
    allocate (efficiency (process%n_channels))
    where (process%grids%grids%f_max /= 0)
       efficiency = process%grids%grids%mu(1) / abs (process%grids%grids%f_max)
    elsewhere
       efficiency = 0
    end where
  end function process_get_vamp_efficiency_array

  function process_get_vamp_efficiency (process) result (efficiency)
    real(default) :: efficiency
    type(process_t), intent(in) :: process
    real(default), dimension(:), allocatable :: weight
    real(default) :: norm
    allocate (weight (process%n_channels))
    weight = process%grids%weights * abs (process%grids%grids%f_max)
    norm = sum (weight)
    if (norm /= 0) then
       efficiency = &
            dot_product (process_get_vamp_efficiency_array (process), weight) &
            / norm
    else
       efficiency = 1
    end if
  end function process_get_vamp_efficiency

  subroutine process_integrate (process, rng, &
       grid_parameters, pass, iterations, calls, &
       discard_integrals, adapt_grids, adapt_weights, print_current)
    type(process_t), intent(inout), target :: process
    type(tao_random_state), intent(inout) :: rng
    type(grid_parameters_t), intent(in) :: grid_parameters
    integer, intent(in) :: pass, iterations, calls
    logical, intent(in) :: discard_integrals
    logical, intent(in) :: adapt_grids
    logical, intent(in) :: adapt_weights
    logical, intent(in) :: print_current
    integer :: it
    real(default) :: integral, error, efficiency
    integer :: u
    u = logfile_unit ()
    if (discard_integrals) then
       call vamp_discard_integrals (process%grids, &
            calls, stratified=grid_parameters%stratified, eq=process%vamp_eq)
    end if
    process%beams_are_set = .false.
    do it = 1, iterations
       if (adapt_grids) then
          call process_adapt_grids (process)
       end if
       if (adapt_weights) then
          call process_adapt_channel_weights (process, grid_parameters, calls)
       end if
       call vamp_sample_grids &
            (rng, process%grids, sample_function, process%store_index, 1, &
             integral=integral, std_dev=error)
       efficiency = process_get_vamp_efficiency (process)
       call integration_results_append (process%results, &
            process%type, pass, 1, calls, &
            integral, error, efficiency)
       if (print_current) then
          call integration_results_write_current (process%results)
          call integration_results_write_current (process%results, unit=u)
       end if
    end do
  end subroutine process_integrate

  subroutine process_adapt_grids (process)
    type(process_t), intent(inout), target :: process
    call vamp_refine_grids (process%grids)
  end subroutine process_adapt_grids

  subroutine process_adapt_channel_weights (process, grid_parameters, calls)
    type(process_t), intent(inout), target :: process
    type(grid_parameters_t), intent(in) :: grid_parameters
    integer, intent(in) :: calls
    real(default), dimension(:), allocatable :: weights
    integer :: g, i0, i1, n
    real(default) :: sum_weights, weight_min
    logical, dimension(:), allocatable :: weight_underflow
    real(default) :: sum_weight_underflow
    integer :: n_underflow
    allocate (weights (process%n_channels))
    weights = process%grids%weights &
         * vamp_get_variance (process%grids%grids) &
           ** grid_parameters%channel_weights_power
    do g = 1, phs_forest_get_n_groves (process%forest)
       call phs_forest_get_grove_bounds (process%forest, g, i0, i1, n)
       weights(i0:i1) = sum (weights(i0:i1)) / n
    end do
    sum_weights = sum (weights)
    if (sum_weights /= 0) then
       weights = weights / sum (weights)
       if (grid_parameters%min_calls_per_channel /= 0) then
          weight_min = real (grid_parameters%min_calls_per_channel, default) &
               / calls
          weight_underflow = weights /= 0 .and. weights < weight_min
          n_underflow = count (weight_underflow)
          sum_weight_underflow = sum (weights, mask=weight_underflow)
          where (weight_underflow)
             weights = weight_min
          elsewhere
             weights = weights &
                  * (1 - n_underflow * weight_min) / (1 - sum_weight_underflow)
          end where
       end if
       call vamp_update_weights (process%grids, weights)
    end if
  end subroutine process_adapt_channel_weights

  subroutine process_setup_event_generation (process, qn_mask_in)
    type(process_t), intent(inout), target :: process
    type(quantum_numbers_mask_t), intent(in), optional :: qn_mask_in
    integer, dimension(:), allocatable :: coll_index
    type(quantum_numbers_mask_t), dimension(:), allocatable :: mask_in
    type(quantum_numbers_mask_t) :: mask_conn_sqme, mask_conn_flows
    type(evaluator_t), pointer :: eval_sfchain, eval_sqme, eval_flows
    type(interaction_t), pointer :: int_hi, int_beam
    integer :: n_in, n_out, n_tot, i
    call hard_interaction_final_sqme (process%hi)
    call hard_interaction_final_flows (process%hi)
    if (process%use_beams) then
       call evaluator_final (process%eval_sqme)
       call evaluator_final (process%eval_flows)
    end if
    n_in  = hard_interaction_get_n_in  (process%hi)
    n_out = hard_interaction_get_n_out (process%hi)
    n_tot = hard_interaction_get_n_tot (process%hi)
    int_hi => hard_interaction_get_int_ptr (process%hi)
    call interaction_reset_momenta (int_hi)
    allocate (mask_in (n_in))
    if (process%use_beams) then
       allocate (coll_index (n_in))
       coll_index = strfun_chain_get_colliding_particles (process%sfchain)
       mask_in = strfun_chain_get_colliding_particles_mask (process%sfchain)
       mask_conn_sqme = new_quantum_numbers_mask (.true., .true., .true.)
       mask_conn_flows = new_quantum_numbers_mask (.true., .false., .true.)
    else if (present (qn_mask_in)) then
       mask_in = qn_mask_in
    else
       mask_in = new_quantum_numbers_mask (.false., .false., .true.)
    end if
    call hard_interaction_init_sqme  (process%hi, mask_in)
    call hard_interaction_init_flows (process%hi, mask_in)
    int_beam     => strfun_chain_get_beam_int_ptr       (process%sfchain)
    eval_sfchain => strfun_chain_get_last_evaluator_ptr (process%sfchain)
    eval_sqme    => hard_interaction_get_eval_sqme_ptr  (process%hi)
    eval_flows   => hard_interaction_get_eval_flows_ptr (process%hi)
!     print *, "Hard interaction"
!     call interaction_write (int_hi)
!     print *, "Evaluator: HI: SQME"
!     call evaluator_write (eval_sqme)
!     print *, "Evaluator: HI: flows"
!     call evaluator_write (eval_flows)
    if (process%use_beams) then
       if (associated (eval_sfchain)) then
          do i = 1, n_in
             call evaluator_set_source_link &
                  (eval_sqme,  i, eval_sfchain, coll_index(i))
             call evaluator_set_source_link &
                  (eval_flows, i, eval_sfchain, coll_index(i))
          end do
          call evaluator_init_product &
               (process%eval_sqme,  eval_sfchain, eval_sqme,  mask_conn_sqme)
          call evaluator_init_product &
               (process%eval_flows, eval_sfchain, eval_flows, mask_conn_flows)
       else
          do i = 1, n_in
             call evaluator_set_source_link &
                  (eval_sqme,  i, int_beam, coll_index(i))
             call evaluator_set_source_link &
                  (eval_flows, i, int_beam, coll_index(i))
          end do
          call evaluator_init_product &
               (process%eval_sqme,  int_beam, eval_sqme,  mask_conn_sqme)
          call evaluator_init_product &
               (process%eval_flows, int_beam, eval_flows, mask_conn_flows)
       end if
!        print *, "Evaluator: Beams+HI: SQME"
!        call evaluator_write (process%eval_sqme)
!        print *, "Evaluator: Beams+HI: flows"
!        call evaluator_write (process%eval_flows)
    end if
  end subroutine process_setup_event_generation

  subroutine process_generate_weighted_event (process, rng, channel, weight)
    type(process_t), intent(inout), target :: process
    type(tao_random_state), intent(inout) :: rng
    integer, intent(in) :: channel
    real(default), intent(out) :: weight
    real(default), dimension(process%n_par) :: x
    call vamp_next_event &
         (x, rng, process%grids%grids(channel), &
          sample_function, process%store_index, &
          weight, channel, process%grids%weights, process%grids%grids)
    call process_complete_evaluators (process)
!    call process_do_analysis (process, weight)
  end subroutine process_generate_weighted_event
    
  subroutine process_generate_unweighted_event (process, rng, excess)
    type(process_t), intent(inout), target :: process
    type(tao_random_state), intent(inout) :: rng
    real(default), intent(out), optional :: excess
    real(default), dimension(process%n_par) :: x
    call vamp_next_event &
         (x, rng, process%grids, &
          sample_function, process%store_index, phi_trivial, &
          excess=excess)
    call process_complete_evaluators (process)
    call process_do_analysis (process)
  end subroutine process_generate_unweighted_event

  function phi_trivial (xi, channel_dummy) result (x)
    real(default), dimension(:), intent(in) :: xi
    integer, intent(in) :: channel_dummy
    real(default), dimension(size(xi)) :: x
    x = xi
  end function phi_trivial

  subroutine process_complete_evaluators (process)
    type(process_t), intent(inout), target :: process
    call hard_interaction_evaluate_sqme (process%hi)
    call hard_interaction_evaluate_flows (process%hi)
    if (process%use_beams) then
       call evaluator_receive_momenta (process%eval_sqme)
       call evaluator_receive_momenta (process%eval_flows)
       call evaluator_evaluate (process%eval_sqme)
       call evaluator_evaluate (process%eval_flows)
    end if
  end subroutine process_complete_evaluators

  subroutine process_results_write_header (process, unit, logfile)
    type(process_t), intent(in) :: process
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: logfile
    call write_dline (unit)
    call write_header (process%type, unit, logfile)
    call write_dline (unit)
  end subroutine process_results_write_header

  subroutine process_results_write_current (process, unit)
    type(process_t), intent(in) :: process
    integer, intent(in), optional :: unit
    call integration_results_write_current (process%results, unit)
  end subroutine process_results_write_current

  subroutine process_results_write_current_average (process, unit)
    type(process_t), intent(in) :: process
    integer, intent(in), optional :: unit
    call write_hline (unit)
    call integration_results_write_current_average (process%results, unit)
    call write_hline (unit)
  end subroutine process_results_write_current_average

  subroutine process_results_write_footer (process, unit)
    type(process_t), intent(in) :: process
    integer, intent(in), optional :: unit
    call write_dline (unit)
    call integration_results_write_current_average (process%results, unit)
    call write_dline (unit)
  end subroutine process_results_write_footer
    
  subroutine process_results_write (process, unit)
    type(process_t), intent(in) :: process
    integer, intent(in), optional :: unit
    call integration_results_write (process%results, unit)
  end subroutine process_results_write

  subroutine process_record_integral (process)
    type(process_t), intent(inout) :: process
    call process_library_record_integral &
         (process%prc_lib, process%id, &
          integration_results_get_n_calls (process%results), &
          integration_results_get_integral (process%results), &
          integration_results_get_error (process%results), &
          integration_results_get_accuracy (process%results), &
          integration_results_get_chi2 (process%results), &
          integration_results_get_efficiency (process%results))
  end subroutine process_record_integral
    
  subroutine process_make_copy (process, original)
    type(process_t), intent(inout), target :: process
    type(process_t), intent(in), target, optional :: original
    type(process_t), pointer :: copy
    type(interaction_t), pointer :: beam_int, copy_beam_int
    type(interaction_t), pointer :: hi_int, copy_hi_int
    type(evaluator_t), pointer :: hi_eval_trace, copy_hi_eval_trace
    type(evaluator_t), pointer :: hi_eval_sqme, copy_hi_eval_sqme
    type(evaluator_t), pointer :: hi_eval_flows, copy_hi_eval_flows
    allocate (copy)
    copy%type = process%type
    copy%is_original = .false.
    if (present (original)) then
       copy%original => original
    else
       copy%original => process
    end if
    copy%initialized = process%initialized
    copy%use_beams = process%use_beams
    copy%id = process%id
    copy%prc_lib => process%prc_lib
    copy%lib_index = process%lib_index
    copy%store_index = process%store_index
    copy%model => process%model
    if (process%use_beams) then
       copy%n_strfun = process%n_strfun
       copy%n_par_strfun = process%n_par_strfun
    end if
    copy%n_par_hi = process%n_par_hi
    copy%n_par = process%n_par
    copy%azimuthal_dependence = process%azimuthal_dependence
    copy%vamp_grids_defined = process%vamp_grids_defined
    copy%sqrts_known = process%sqrts_known
    copy%sqrts = process%sqrts
    if (process%use_beams) then
       if (allocated (process%x_strfun)) &
            allocate (copy%x_strfun (size (process%x_strfun)))
    end if
    if (allocated (process%x_hi)) &
         allocate (copy%x_hi (size (process%x_hi)))
    copy%n_channels = process%n_channels
    if (allocated (process%x)) &
         allocate (copy%x (size (process%x, 1), size (process%x, 2)))
    if (allocated (process%phs_factor)) &
         allocate (copy%phs_factor (size (process%phs_factor)))
    if (allocated (process%mass_in)) then
       allocate (copy%mass_in (size (process%mass_in)))
       copy%mass_in = process%mass_in
    end if
    copy%averaging_factor = process%averaging_factor
    if (process%use_beams)  copy%sfchain = process%sfchain
    copy%hi = process%hi
    if (process%use_beams)  copy%eval_trace = process%eval_trace
    !  copy%eval_sqme = process%eval_sqme
    !  copy%eval_flows = process%eval_flows
    copy%forest = process%forest
    copy%vamp_eq = process%vamp_eq
    copy%prt_list = process%prt_list
    copy%var_list = process%var_list
    copy%cut_expr = process%cut_expr
    copy%reweighting_expr = process%reweighting_expr
    copy%scale_expr = process%scale_expr
    if (allocated (process%active_channel)) then
       allocate (copy%active_channel (size (process%active_channel)))
       copy%active_channel = process%active_channel
    end if
    call vamp_copy_grids (copy%grids, process%grids)
    beam_int => strfun_chain_get_beam_int_ptr (process%sfchain)
    hi_int => hard_interaction_get_int_ptr (process%hi)
    hi_eval_trace => hard_interaction_get_eval_trace_ptr (process%hi)
    hi_eval_sqme => hard_interaction_get_eval_sqme_ptr (process%hi)
    hi_eval_flows => hard_interaction_get_eval_flows_ptr (process%hi)
    copy_beam_int => strfun_chain_get_beam_int_ptr (copy%sfchain)
    copy_hi_int => hard_interaction_get_int_ptr (copy%hi)
    copy_hi_eval_trace => hard_interaction_get_eval_trace_ptr (copy%hi)
    copy_hi_eval_sqme => hard_interaction_get_eval_sqme_ptr (copy%hi)
    copy_hi_eval_flows => hard_interaction_get_eval_flows_ptr (copy%hi)
    select case (process%type)
    case (PRC_SCATTERING)
       call msg_bug ("Process copy for scattering processes not implemented")
    case (PRC_DECAY)
       call interaction_reassign_links &
            (copy_hi_int, beam_int, copy_beam_int)
       call evaluator_reassign_links &
            (copy_hi_eval_trace, beam_int, copy_beam_int)
       call evaluator_reassign_links &
            (copy_hi_eval_sqme, beam_int, copy_beam_int)
       call evaluator_reassign_links &
            (copy_hi_eval_flows, beam_int, copy_beam_int)
       call evaluator_reassign_links &
            (copy_hi_eval_trace, hi_int, copy_hi_int)
       call evaluator_reassign_links &
            (copy_hi_eval_sqme, hi_int, copy_hi_int)
       call evaluator_reassign_links &
            (copy_hi_eval_flows, hi_int, copy_hi_int)
       call evaluator_reassign_links &
            (copy%eval_trace, beam_int, copy_beam_int)
       call evaluator_reassign_links &
            (copy%eval_sqme, beam_int, copy_beam_int)
       call evaluator_reassign_links &
            (copy%eval_flows, beam_int, copy_beam_int)
       call evaluator_reassign_links &
            (copy%eval_trace, hi_eval_trace, copy_hi_eval_trace)
       call evaluator_reassign_links &
            (copy%eval_sqme, hi_eval_sqme, copy_hi_eval_sqme)
       call evaluator_reassign_links &
            (copy%eval_flows, hi_eval_flows, copy_hi_eval_flows)
    end select
    process%copy => copy
  end subroutine process_make_copy

  recursive subroutine process_request_copy (process, copy, original)
    type(process_t), intent(inout), target :: process
    type(process_t), pointer :: copy
    type(process_t), intent(inout), target, optional :: original
    if (associated (process%copy)) then
       if (process%copy%in_use) then
          call process_request_copy (process%copy, copy, process)
       else
          copy => process%copy
          copy%in_use = .true.
          if (present (original)) then
             original%working_copy => copy
          else
             process%working_copy => copy
          end if
       end if
    else
       call process_make_copy (process, original)
       copy => process%copy
       copy%in_use = .true.
       if (present (original)) then
          original%working_copy => copy
       else
          process%working_copy => copy
       end if
    end if
  end subroutine process_request_copy

  function process_get_working_copy_ptr (process) result (copy)
    type(process_t), intent(in), target :: process 
    type(process_t), pointer :: copy
    if (associated (process%working_copy)) then
       copy => process%working_copy
    else
       copy => process
    end if
  end function process_get_working_copy_ptr

  subroutine process_copy_free (process)
    type(process_t), intent(inout), target :: process
    process%in_use = .false.
    if (associated (process%original)) then
       process%original%working_copy => null ()
    end if
  end subroutine process_copy_free

  subroutine process_store_final ()
    type(process_entry_t), pointer :: current
    if (allocated (store%proc))  deallocate (store%proc)
    store%last => null ()
    do while (associated (store%first))
       current => store%first
       store%first => current%next
       call process_final (current%process)
       deallocate (current)
    end do
    store%n = 0
  end subroutine process_store_final

  subroutine process_store_write (unit)
    integer, intent(in), optional :: unit
    type(process_t), pointer :: process
    integer :: u, i
    u = output_unit (unit);  if (u < 0)  return
    write (u, *)  repeat ("%", 78)
    write (u, *)  "Process store contents"
    do i = 1, store%n
       write (u, *)  repeat ("%", 78)
       write (u, *)  "Process No.", i
       process => store%proc(i)%ptr
       call process_write (process, unit)
    end do
    write (u, *)  "Process store end"
    write (u, *)  repeat ("%", 78)
  end subroutine process_store_write

  subroutine process_store_write_results (unit)
    integer, intent(in), optional :: unit
    type(process_t), pointer :: process
    type(string_t), dimension(:), allocatable :: process_id
    real(default), dimension(:), allocatable :: integral, error
    type(string_t), dimension(:), allocatable :: phys_unit
    integer :: u, i, process_id_len
    character(12) :: fmt
    u = output_unit (unit);  if (u < 0)  return
    allocate (process_id (store%n), phys_unit (store%n))
    allocate (integral (store%n), error (store%n))
    do i = 1, store%n
       process => store%proc(i)%ptr
       if (process%initialized) then
          process_id(i) = process%id
          integral(i) = process_get_integral (process)
          error(i) = process_get_error (process)
          select case (process%type)
          case (PRC_DECAY);       phys_unit(i) = "GeV"
          case (PRC_SCATTERING);  phys_unit(i) = "fb"
          case default;           phys_unit(i) = "[undefined]"
          end select
       else
          process_id(i) = ""
       end if
    end do
    write (u, "(A)")  "|=========================      Results Summary      =========================|"
    if (store%n == 0) then
       write (u, *) "[empty]"
    else
       process_id_len = maxval (len (process_id))
       write (fmt, "(A,I0,A)")  "(1x,A", process_id_len + 1, ")"
       do i = 1, store%n
          if (process_id(i) /= "") then
             write (u, fmt, advance="no")  char (process_id(i)) // ":"
             write (u, "(1x, 1PE15.8, 1x, '+-', 1x, 1PE8.2)", advance="no") &
                  integral(i), error(i)
             write (u, "(1x, A)")  char (phys_unit(i))
          end if
       end do
    end if
    write (u, "(A)")  "|=============================================================================|"
  end subroutine process_store_write_results

  function process_store_get_n_processes () result (n)
    integer :: n
    n = store%n
  end function process_store_get_n_processes

  function process_store_get_entry_ptr (process_id) result (entry)
    type(process_entry_t), pointer :: entry
    type(string_t), intent(in) :: process_id
    entry => store%first
    do while (associated (entry))
       if (entry%process%id == process_id)  exit
       entry => entry%next
    end do
  end function process_store_get_entry_ptr

  function process_store_get_process_index (process_id) result (process_index)
    integer :: process_index
    type(string_t), intent(in) :: process_id
    type(process_entry_t), pointer :: entry
    entry => process_store_get_entry_ptr (process_id)
    if (associated (entry)) then
       process_index = entry%process%store_index
    else
       process_index = 0
    end if
  end function process_store_get_process_index

  function process_store_get_process_ptr_int (i) result (process)
    type(process_t), pointer :: process
    integer, intent(in) :: i
    process => store%proc(i)%ptr
  end function process_store_get_process_ptr_int

  function process_store_get_process_ptr_id (id) result (process)
    type(process_t), pointer :: process
    type(string_t), intent(in) :: id
    integer :: i
    do i = 1, store%n
       process => store%proc(i)%ptr
       if (process%id == id)  return
    end do
    process => null ()
  end function process_store_get_process_ptr_id

  function process_store_get_fresh_process_ptr (process_id) result (process)
    type(process_t), pointer :: process
    type(string_t), intent(in) :: process_id
    type(process_entry_t), pointer :: current, entry
    integer :: i
    integer, parameter :: BLOCK_SIZE = 10
    current => process_store_get_entry_ptr (process_id)
    if (associated (current)) then
       call process_final (current%process)
    else
       allocate (current)
       if (store%n == 0) then
          allocate (store%proc (BLOCK_SIZE))
          store%first => current
       else
          store%last%next => current
       end if
       store%last => current
       store%n = store%n + 1
       if (store%n <= size (store%proc)) then
          store%proc(store%n)%ptr => current%process
       else
          deallocate (store%proc)
          allocate (store%proc (store%n + BLOCK_SIZE))
          i = 1
          entry => store%first
          do while (associated (entry))
             store%proc(i)%ptr => entry%process;  i = i + 1
             entry => entry%next
          end do
       end if
    end if
    process => current%process
  end function process_store_get_fresh_process_ptr

  subroutine process_store_init_process &
       (process, prc_lib, process_id, model, var_list, use_beams)
    type(process_t), pointer :: process
    type(process_library_t), intent(in), target :: prc_lib
    type(string_t), intent(in) :: process_id
    type(model_t), intent(in), target :: model
    type(var_list_t), intent(in), optional, target :: var_list
    logical, intent(in), optional :: use_beams
    integer :: process_lib_index, process_store_index
    process_lib_index = process_library_get_process_index (prc_lib, process_id)
    if (process_lib_index == 0) then
       call msg_fatal ("Process '" // char (process_id) &
            // "' is not available.")
    end if
    process_store_index = process_store_get_process_index (process_id)
    process => process_store_get_fresh_process_ptr (process_id)
    if (process_store_index == 0) then
       process_store_index = process_store_get_n_processes ()
    end if
    call process_init &
         (process, prc_lib, process_lib_index, process_store_index, &
          process_id, model, var_list, use_beams)
  end subroutine process_store_init_process

  function sample_function (xi, prc_index, weights, channel, grids) result (f)
    real(default) :: f
    real(default), dimension(:), intent(in) :: xi
    integer, intent(in) :: prc_index
    real(default), dimension(:), intent(in), optional :: weights
    integer, intent(in), optional :: channel
    type(vamp_grid), dimension(:), intent(in), optional :: grids
    type(process_t), pointer :: process
    logical :: ok
    process => process_get_working_copy_ptr (store%proc(prc_index)%ptr)
    call process_set_kinematics (process, xi, channel, ok)
    if (ok)  ok = process_passes_cuts (process)
    if (ok) then
       call process_compute_vamp_phs_factor (process, weights)
       call process_evaluate (process)
       call process_compute_reweighting_factor (process)
       process%sample_function_value = &
            process%flux_factor &
            * process%sf_mapping_factor &
            * process%vamp_phs_factor &
            * process%phs_volume &
            * process%sqme &
            * process%reweighting_factor
    else
       process%sample_function_value = 0
    end if
    f = process%sample_function_value
  end function sample_function

  subroutine process_test ()
    type(os_data_t) :: os_data
    type(process_library_t) :: prc_lib
    type(model_t), pointer :: model
    print *, "*** Load process library"
    call process_library_init (prc_lib, var_str("proc"), os_data)
    call process_library_load (prc_lib, os_data)
    print *
    print *, "*** Read model file"
    call syntax_model_file_init ()
    call model_list_read_model &
         (var_str("QCD"), var_str("test.mdl"), var_str("./"), model)
    call syntax_pexpr_init ()
    call syntax_phs_forest_init ()
    print *
    call process_test1 (prc_lib, model)
    print *
    call process_test2 (prc_lib, model)
    print *
    call process_test3 (prc_lib, model)
!     print *
!     call process_test4 (prc_lib, model)
    print *
    print *, "* Cleanup"
    call process_store_final ()
    call syntax_pexpr_final ()
    call syntax_phs_forest_final ()
    call syntax_model_file_final ()
    call process_library_final (prc_lib)
  end subroutine process_test

  subroutine process_test1 (prc_lib, model)
    type(process_library_t), intent(in) :: prc_lib
    type(model_t), intent(in), target :: model
    type(process_t), pointer :: process
    type(phs_parameters_t) :: phs_par
    type(mapping_defaults_t) :: mapping_defaults
    type(flavor_t), dimension(1) :: flv
    type(polarization_t), dimension(1) :: pol
    type(beam_data_t) :: beam_data
    type(grid_parameters_t) :: grid_parameters
    type(tao_random_state) :: rng
    integer :: i
    logical :: discard_integrals, adapt_grids, adapt_weights, print_current
    print *, "*** Test decay process"
    print *
    print *, "* Initialization"
    call tao_random_create (rng, 0)
    call process_store_init_process (process, prc_lib, var_str ("zee"), model)
    print *, "  Process ID = ", char (process%id)
    print *
    print *, "*** Beam/strfun setup (unpolarized)"
    print *
    call flavor_init (flv, (/ 23 /), model)
    call polarization_init_unpolarized (pol(1), flv(1))
    call beam_data_init_decay (beam_data, flv, pol)
    call process_setup_beams (process, beam_data, 0, 0)
    call process_connect_strfun (process)
    print *
    print *, "* Phase space setup"
    call process_setup_phase_space (process, &
         phs_par, mapping_defaults, filename_out=var_str("zee.phs"))
    print *
    print *, "*** Test integration"
    print *, "* Grids setup"
    grid_parameters%stratified = .false.
    call process_setup_grids (process, grid_parameters, calls=1000)
    print *
    print *, "* 1 iteration with minimal number of calls"
    call process_results_write_header (process)
    do i = 1, 1
       discard_integrals = i==1
       adapt_grids = .true.
       adapt_weights = .true.
       print_current = .true.
       call process_integrate (process, rng, grid_parameters, &
            1, 1, 9, &
            discard_integrals, adapt_grids, adapt_weights, print_current)
    end do
    call process_results_write_footer (process)
    print *
    print *, "* Process written to 'fort.60'"
    call process_write (process, 60)
    print *
    print *, "*** Beam/strfun setup (polarized)"
    call process_store_init_process (process, prc_lib, var_str ("zee"), model)
    call flavor_init (flv, (/ 23 /), model)
    call polarization_init_axis &
         (pol(1), flv(1), (/ 0._default, 0._default, 1._default/))
    call beam_data_init_decay (beam_data, flv, pol)
    call process_setup_beams (process, beam_data, 0, 0)
    call process_connect_strfun (process)
    print *
    print *, "* Phase space setup"
    call process_setup_phase_space (process, &
         phs_par, mapping_defaults, filename_out=var_str("zee.phs"))
    print *
    print *, "*** Test integration"
    print *, "* Grids setup"
    grid_parameters%stratified = .false.
    call process_setup_grids (process, grid_parameters, calls=1000)
    print *
    print *, "* 3 + 3 iterations"
    call process_results_write_header (process)
    do i = 1, 3
       discard_integrals = i==1
       adapt_grids = .true.
       adapt_weights = .true.
       print_current = .true.
       call process_integrate (process, rng, grid_parameters, &
            1, 1, 10000, &
            discard_integrals, adapt_grids, adapt_weights, print_current)
    end do
    call process_results_write_current_average (process)
    call process_integrate (process, rng, grid_parameters, &
         2, 3, 10000, &
         .true., .true., .false., .true.)
    call process_results_write_footer (process)
    print *
    print *, "* Process written to 'fort.61'"
    call process_write (process, 61)
  end subroutine process_test1

  subroutine process_test2 (prc_lib, model)
    type(process_library_t), intent(in) :: prc_lib
    type(model_t), intent(in), target :: model
    type(process_t), pointer :: process
    type(phs_parameters_t) :: phs_par
    type(mapping_defaults_t) :: mapping_defaults
    type(flavor_t), dimension(1) :: flv
    type(polarization_t), dimension(1) :: pol
    type(beam_data_t) :: beam_data
    type(grid_parameters_t) :: grid_parameters
    type(tao_random_state) :: rng
    real(default) :: weight
    integer :: i
    logical :: discard_integrals, adapt_grids, adapt_weights, print_current
    print *, "*** Test decay process"
    print *
    print *, "* Initialization"
    call tao_random_create (rng, 0)
    call process_store_init_process &
         (process, prc_lib, var_str ("zqq"), model, use_beams=.false.)
    print *, "  Process ID = ", char (process%id)
    print *
    print *, "*** Beam/strfun setup (unpolarized)"
    print *
    call flavor_init (flv, (/ 23 /), model)
    call polarization_init_unpolarized (pol(1), flv(1))
    call beam_data_init_decay (beam_data, flv, pol)
    call process_setup_beams (process, beam_data, 0, 0)
    call process_connect_strfun (process)
    print *
    print *, "* Phase space setup"
    call process_setup_phase_space (process, &
         phs_par, mapping_defaults, filename_out=var_str("zee.phs"))
    print *
    print *, "*** Test integration"
    print *, "* Grids setup"
    grid_parameters%stratified = .false.
    call process_setup_grids (process, grid_parameters, calls=1000)
    print *
    print *, "* 1 iteration with minimal number of calls"
    call process_results_write_header (process)
    do i = 1, 1
       discard_integrals = i==1
       adapt_grids = .true.
       adapt_weights = .true.
       print_current = .true.
       call process_integrate (process, rng, grid_parameters, &
            1, 1, 9, &
            discard_integrals, adapt_grids, adapt_weights, print_current)
    end do
    call process_results_write_footer (process)
    print *
    print *, "* Process written to 'fort.62'"
    call process_write (process, 62)
    print *
    print *, "*** Event generation"
    call process_setup_event_generation (process)
    print *
    print *, "* Generate weighted event"
    call process_generate_weighted_event (process, rng, 1, weight)
    print *
    print *, "* Process written to 'fort.63'"
    call process_write (process, 63)
    print *, "weight =", weight
    print *
    print *, "* Generate unweighted event"
    call process_generate_unweighted_event (process, rng, weight)
    print *
    print *, "* Process written to 'fort.64'"
    call process_write (process, 64)
    print *, "excess weight =", weight
  end subroutine process_test2

  subroutine process_test3 (prc_lib, model)
    type(process_library_t), intent(in) :: prc_lib
    type(model_t), intent(in), target :: model
    type(process_t), pointer :: process
    type(phs_parameters_t) :: phs_par
    type(mapping_defaults_t) :: mapping_defaults
    type(flavor_t), dimension(2) :: flv
    type(polarization_t), dimension(2) :: pol
    type(beam_data_t) :: beam_data
    type(grid_parameters_t) :: grid_parameters
    real(default), dimension(:), allocatable :: x
    integer :: channel
    logical :: ok
    integer :: i
    type(tao_random_state) :: rng
    logical :: discard_integrals, adapt_grids, adapt_weights, print_current
    print *, "*** Test scattering process"
    print *
    print *, "* Initialization"
    call tao_random_create (rng, 0)
    call process_store_init_process (process, prc_lib, var_str ("nnh"), model)
    print *, "  Process ID = ", char (process%id)
    print *
    print *, "* Beam/strfun setup"
    print *
    call flavor_init (flv, (/ 11, -11 /), model)
    call polarization_init_unpolarized (pol(1), flv(1))
    call polarization_init_unpolarized (pol(2), flv(2))
    call beam_data_init_sqrts (beam_data, 500._default, flv, pol)
    call process_setup_beams (process, beam_data, 0, 0)
    call process_connect_strfun (process)
    print *
    print *, "* Phase space setup"
    call process_setup_phase_space (process, &
         phs_par, mapping_defaults, filename_out=var_str("nnh.phs"))
    print *
    print *, "* Kinematics setup"
    allocate (x (process_get_n_parameters (process)))
    do i = 1, size (x)
       x(i) = (i - 0.5_default) * (1._default / size (x))
    end do
    channel = 1
    call process_set_kinematics (process, x, channel, ok)
    print *
    print *, "* Process written to 'fort.70'"
    call process_write (process, 70)
    print *
    print *, "*** Test process evaluation"
    call process_evaluate (process)
    print *
    print *, "* Process written to 'fort.71'"
    call process_write (process, 71)
    print *
    print *, "*** Test integration"
    print *, "* Grids setup"
    call process_setup_grids (process, grid_parameters, calls=10000)
    print *
    print *, "* 5 + 3 iterations"
    call process_results_write_header (process)
    do i = 1, 5
       discard_integrals = i==1
       adapt_grids = .true.
       adapt_weights = .true.
       print_current = .true.
       call process_integrate (process, rng, grid_parameters, &
            1, 1, 10000, &
            discard_integrals, adapt_grids, adapt_weights, print_current)
    end do
    call process_results_write_current_average (process)
    call process_integrate (process, rng, grid_parameters, &
         2, 3, 20000, .true., .false., .false., .true.)
    call process_results_write_footer (process)
    print *
    print *, "* Process written to 'fort.72'"
    call process_write (process, 72)
  end subroutine process_test3

  subroutine process_test4 (prc_lib, model)
    type(process_library_t), intent(in) :: prc_lib
    type(model_t), intent(in), target :: model
    type(process_t), pointer :: process
    type(phs_parameters_t) :: phs_par
    type(mapping_defaults_t) :: mapping_defaults
    type(flavor_t), dimension(2) :: flv
    type(polarization_t), dimension(2) :: pol
    type(beam_data_t) :: beam_data
    type(lhapdf_data_t), dimension(2) :: data
    type(stream_t) :: stream
    type(parse_tree_t) :: parse_tree
    type(grid_parameters_t) :: grid_parameters
    integer :: i
    type(tao_random_state) :: rng
    print *, "*** Test process setup"
    print *
    print *, "* Initialization"
    call tao_random_create (rng, 0)
    call process_store_init_process (process, prc_lib, var_str ("qq"), model)
    print *, "  Process ID = ", char (process%id)
    print *
    print *, "* Beam/strfun setup"
    print *
!    call flavor_init (flv, (/ 21, 21 /), model)
    call flavor_init (flv, (/ PROTON, PROTON /), model)
    call polarization_init_unpolarized (pol(1), flv(1))
    call polarization_init_unpolarized (pol(2), flv(2))
    call beam_data_init_sqrts (beam_data, 14000._default, flv, pol)
!     call process_setup_beams (process, beam_data, 0, 0)
    call process_setup_beams (process, beam_data, 2, 0)
    call lhapdf_init ()
    call lhapdf_data_init (data(1), model, flv(1))
    call lhapdf_data_init (data(2), model, flv(2))
    call process_set_strfun (process, 1, 1, data(1), 1) 
    call process_set_strfun (process, 2, 2, data(2), 1)
    call process_connect_strfun (process)
    print *
    print *, "* Phase space setup"
    call process_setup_phase_space (process, &
         phs_par, mapping_defaults, filename_out=var_str("qq.phs"))
    print *
    print *, "* Cuts setup"
    call stream_init (stream, var_str ("all Pt > 50 GeV (outgoing u:d:U:D)"))
    call parse_tree_init_lexpr (parse_tree, stream, .true.)
    call process_setup_cuts (process, parse_tree_get_root_ptr (parse_tree))
    call parse_tree_final (parse_tree)
    call stream_final (stream)
    print *
    print *, "* Scale setup"
    call stream_init (stream, var_str ("1 TeV"))
    call parse_tree_init_expr (parse_tree, stream, .true.)
    call process_setup_scale (process, parse_tree_get_root_ptr (parse_tree))
    call parse_tree_final (parse_tree)
    call stream_final (stream)
    print *
    print *, "*** Test integration"
    print *, "* Grids setup"
    call process_setup_grids (process, grid_parameters, calls=10000)
    print *
    print *, "* 15 + 3 iterations"
    call process_results_write_header (process)
    do i = 1, 15
       call process_integrate (process, rng, grid_parameters, &
            1, 1, 50000, i==1, .true., i>2, .true.)
    end do
    call process_results_write_current_average (process)
    call process_integrate (process, rng, grid_parameters, &
         2, 3, 50000, .true., .false., .true., .true.)
    call process_results_write_footer (process)
    print *
    print *, "* Process written to 'fort.90'"
    call process_write (process, 90)
  end subroutine process_test4


end module processes
